import argparse
import time
import os

from concurrent.futures import ProcessPoolExecutor

from lazurite import util
import lazurite.project.project
from lazurite.material import Material
from lazurite.material.stage import ShaderStage
from lazurite.material.platform import ShaderPlatform
from lazurite.material.encryption import EncryptionType
from lazurite.compiler.macro_define import MacroDefine


def list_packed_materials(args) -> list[str]:
    material_files = []
    if len(args.inputs) == 0:
        raise Exception("Material path argument was expected")

    for material_path in args.inputs:
        if not os.path.exists(material_path):
            raise Exception("Invalid path to material or folder")

        if os.path.isfile(material_path):
            material_files.append(material_path)
        else:
            for entry in os.listdir(material_path):
                entry: str
                file_path = os.path.join(material_path, entry)
                if os.path.isfile(file_path) and file_path.endswith(Material.EXTENSION):
                    material_files.append(file_path)

    return material_files


def list_unpacked_materials(args):
    material_folders = []
    for path in args.inputs:
        if not os.path.exists(path):
            raise Exception("Invalid path to material or folder")

        if os.path.isfile(path) and os.path.basename(path) == "material.json":
            material_folders.append(os.path.dirname(path))
        elif os.path.isdir(path) and os.path.isfile(
            os.path.join(path, "material.json")
        ):
            material_folders.append(path)
        elif os.path.isdir(path):
            for child_path in os.listdir(path):
                child_path = os.path.join(path, child_path)
                if os.path.isdir(child_path) and os.path.isfile(
                    os.path.join(child_path, "material.json")
                ):
                    material_folders.append(child_path)

    return material_folders


def unpack(args):
    for file in list_packed_materials(args):
        file_name: str = os.path.basename(file)
        print(file_name)

        material = Material.load_bin_file(file)
        if args.sort_flags:
            material.sort_variants()
        material.store(
            file_name.removesuffix(Material.EXTENSION), args.output, args.skip_shaders
        )


def pack(args):
    for file in list_unpacked_materials(args):
        file_name: str = os.path.basename(os.path.abspath(file)) + Material.EXTENSION
        print(file_name)

        material = Material()
        material.load_unpacked_material(file)

        with open(os.path.join(args.output, file_name), "wb") as f:
            material.write(f)


def label(args):
    for file in list_packed_materials(args):
        file_name: str = os.path.basename(file)
        print(file_name)

        material = Material.load_bin_file(file)
        material.label()
        with open(os.path.join(args.output, file_name), "wb") as f:
            material.write(f)


def clear(args):
    for file in list_packed_materials(args):
        file_name: str = os.path.basename(file)
        print(file_name)

        material = Material.load_bin_file(file)

        for shader_pass in material.passes:
            for variant in shader_pass.variants:
                for shader in variant.shaders:
                    shader.bgfx_shader.shader_bytes = bytes()

        material.encryption = EncryptionType.NONE
        with open(os.path.join(args.output, file_name), "wb") as f:
            material.write(f)


def restore_single_material(args, file: str):
    file_name: str = os.path.basename(file)
    print(file_name)
    file_name = file_name.removesuffix(Material.EXTENSION)
    material = Material.load_bin_file(file)

    material.passes.sort(key=lambda x: x.name)
    material.sort_variants()
    varying = material.restore_varying_def(args.timeout)
    if varying:
        with open(os.path.join(args.output, file_name + ".varying.def.sc"), "w") as f:
            f.write(varying)
    else:
        print(
            "Failed to generate varying.def.sc file, no input/output definitions were found in the target material."
        )

    # That should keep the project safe legally.
    if material.encryption != EncryptionType.NONE:
        print(
            f"Warning! {file_name} material is encrypted. "
            "This tool cannot be used to restore decrypted shaders."
        )
        return

    shader_codes = material.restore_shaders(
        {ShaderPlatform.ESSL_310, ShaderPlatform.ESSL_300},
        set(ShaderStage),
        args.split_passes,
        args.merge_stages,
        not args.no_processing,
        args.timeout,
    )
    for platform, stage, shader_pass, code in shader_codes:
        file_name_tokens = [file_name]
        if args.split_passes:
            file_name_tokens.append(shader_pass)
        file_name_tokens.append(platform.name)
        if not args.merge_stages:
            file_name_tokens.append(stage.name)
        file_name_tokens.append(
            platform.file_extension() if args.no_processing else "sc"
        )
        with open(
            os.path.join(args.output, ".".join(file_name_tokens)),
            "w",
        ) as f:
            f.write(code)


def restore(args):
    paths = list_packed_materials(args)

    # Roughly sort tasks by complexity (estimated from file size).
    paths = list(zip(paths, (os.path.getsize(p) for p in paths)))
    paths.sort(key=lambda f: f[1], reverse=True)
    paths = [p[0] for p in paths]

    with ProcessPoolExecutor(max_workers=args.max_workers or None) as executor:
        executor.map(restore_single_material, [args] * len(paths), paths)


def build(args):
    defines = [MacroDefine.from_string(d) for d in args.defines]
    for path in args.inputs:
        print(f'Compiling project "{path}"')
        lazurite.project.project.compile(
            path,
            args.profile,
            args.output,
            defines,
            args.shaderc,
            args.dxc,
            args.shaderc_args,
            args.dxc_args,
            args.max_workers or None,
        )


def _format_info(obj: list | dict, depth=0) -> str:
    INDENT = "  "
    SEPARATOR = INDENT + "- "
    txt = "\n" if depth and obj else ""
    if isinstance(obj, list):
        txt += "\n".join(
            (SEPARATOR if depth else "") + _format_info(val, depth + 1) for val in obj
        )
        if obj and depth > 1:
            txt = INDENT + txt.replace("\n", "\n" + INDENT)
    elif isinstance(obj, dict):
        txt += "\n".join(
            (SEPARATOR if depth else "")
            + str(key)
            + (f" ({len(val)})" if isinstance(val, list) else "")
            + ": "
            + _format_info(val, depth + 1)
            for key, val in obj.items()
        )
        if obj and depth > 1:
            txt = INDENT + txt.replace("\n", "\n" + INDENT)
    else:
        return str(obj)

    return txt


def info(args):
    for file in list_packed_materials(args):
        file_name: str = os.path.basename(file)
        material = Material.load_bin_file(file)

        shader_count = 0
        for shader_pass in material.passes:
            for variant in shader_pass.variants:
                shader_count += len(variant.shaders)

        material.buffers.sort(key=lambda x: x.reg1)
        material.sort_variants()
        material.passes.sort(key=lambda x: x.name)
        material.uniforms.sort(key=lambda x: x.name)

        platforms = [p.name for p in material.get_platforms()]
        platforms.sort()
        stages = [p.name for p in material.get_stages()]
        stages.sort()

        info = {
            "Name": material.name,
            "Encryption": material.encryption.name,
            "Parent": material.parent,
            "Total Shaders": shader_count,
            "Platforms": platforms,
            "Stages": stages,
            "Passes": {
                p.name: util.generate_pass_name_macro(p.name) for p in material.passes
            },
            "Flags": {
                key: {v: util.generate_flag_name_macro(key, v) for v in value}
                for key, value in material.get_flag_definitions().items()
            },
            "Buffers": {
                f"{b.precision.name} {b.type.name} {b.name}{' '+ b.texture_format if b.texture_format else ''}": {
                    "Reg1": b.reg1,
                    "Reg2": b.reg2,
                    "Unordered Access": b.unordered_access,
                    "Custom Type Info": (
                        {
                            "Struct": b.custom_type_info.struct,
                            "Size": b.custom_type_info.size,
                        }
                        if b.custom_type_info
                        else ""
                    ),
                }
                for b in material.buffers
            },
            "Uniforms": [
                f"{u.type.name} {u.name}{'['+str(u.count)+']' if u.count > 1 else ''}{(' = [' + ', '.join(str(d) for d in u.default)+']') if u.default else ''}"
                for u in material.uniforms
            ],
        }
        print(f"#### {file_name} ####")
        print(_format_info(info))


def main():
    parser = argparse.ArgumentParser(
        prog="lazurite",
        description="Shader development tool for Minecraft: Bedrock Edition with RenderDragon graphics engine",
        epilog="Epilog",
    )

    commands = {
        "unpack": unpack,
        "pack": pack,
        "label": label,
        "restore": restore,
        "build": build,
        "info": info,
        "clear": clear,
        # "project",  # not implemented
        # "config",  # not implemented
    }

    # Common arguments.
    parser.add_argument("command", choices=commands.keys(), help="Command to use")
    parser.add_argument("inputs", nargs="*", help="List of inputs")
    parser.add_argument("-o", "--output", type=str, default="", help="Output path")
    parser.add_argument(
        "--max-workers",
        type=int,
        default=0,
        help="Maximum numbers of cores or processes to use during restore or build commands",
    )

    # Unpack arguments.
    parser.add_argument(
        "--sort-flags",
        action="store_true",
        help="Sorts variants and flags alphabeticaly during unpacking",
    )
    parser.add_argument(
        "--skip-shaders", action="store_true", help="Don't unpack compiled shaders"
    )

    # Restore arguments.
    parser.add_argument(
        "--merge-stages",
        action="store_true",
        help="Restore shader stages in a single file",
    )
    parser.add_argument(
        "--split-passes",
        action="store_true",
        help="Restore separate files for individual passes",
    )
    parser.add_argument(
        "--no-processing",
        action="store_true",
        help="Disable additional processing used for converting from restored GLSL to BGFX SC",
    )
    parser.add_argument(
        "--timeout",
        type=float,
        default=10,
        help="Maximum time allowed for slow search restoring algorithm, in seconds",
    )
    # Not implemented.
    # cli_parser.add_argument("--stages", default=["all"], nargs="*")
    # cli_parser.add_argument("--platforms", default=["essl_310"], nargs="*")

    # Build arguments.
    parser.add_argument(
        "-p",
        "--profile",
        type=str,
        nargs="*",
        help="Profiles to use when building projects",
    )
    parser.add_argument(
        "-d",
        "--defines",
        type=str,
        nargs="*",
        default=[],
        help="Additional defines, passed to shader compiler",
    )
    parser.add_argument("--dxc", type=str, default=None, help="DXC compiler command")
    parser.add_argument(
        "--shaderc", type=str, default=None, help="SHADERC compiler command"
    )
    parser.add_argument(
        "--shaderc-args",
        type=str,
        nargs="*",
        default=[],
        help="Additional DXC compiler arguments",
    )
    parser.add_argument(
        "--dxc-args",
        type=str,
        nargs="*",
        default=[],
        help="Additional SHADERC compiler arguments",
    )

    # Execute command.
    args = parser.parse_args()
    current_time = time.perf_counter()

    commands[args.command](args)

    print(f"Completed in {round(time.perf_counter() - current_time, 2)} seconds")



class MacroDefine:
    name: str
    value: int | None

    def __init__(self, name: str = "", value: int = None) -> None:
        self.name = name
        self.value = value

    def format_bgfx(self):
        if self.value is None:
            return self.name
        return f"{self.name}={self.value}"

    def format_dxc(self):
        if self.value is None:
            return self.name
        return f"{self.name}={self.value}"

    def format_cpp(self):
        if self.value is None:
            return self.name
        return f"{self.name} {self.value}"

    @classmethod
    def from_string(cls, string: str):
        elements = string.split(" ", 1)
        return cls(elements[0], elements[1] if len(elements) > 1 else None)



from lazurite.material.stage import ShaderStage


class ShaderFileOverwrite:
    entry_point: str
    fragment: str
    vertex: str
    compute: str
    unknown: str
    varying: str

    def __init__(self) -> None:
        self.entry_point = ""
        self.fragment = "shaders/fragment.sc"
        self.vertex = "shaders/vertex.sc"
        self.compute = "shaders/compute.sc"
        self.unknown = "shaders/unknown.sc"
        self.varying = "shaders/varying.def.sc"

    def read_json(self, json_data: dict):
        self.entry_point = json_data.get("entry_point", self.entry_point)
        self.fragment = json_data.get("fragment", self.fragment)
        self.vertex = json_data.get("vertex", self.vertex)
        self.compute = json_data.get("compute", self.compute)
        self.unknown = json_data.get("unknown", self.unknown)
        self.varying = json_data.get("varying", self.varying)

    def get_stage(self, stage: ShaderStage):
        if stage == ShaderStage.Fragment:
            return self.fragment
        if stage == ShaderStage.Vertex:
            return self.vertex
        if stage == ShaderStage.Compute:
            return self.compute
        if stage == ShaderStage.Unknown:
            return self.unknown



import re


from .macro_decompiler import restore_code, Variant
from lazurite import util
from lazurite.material.stage import ShaderStage
from lazurite.material.platform import ShaderPlatform
from lazurite.material.shader_pass.shader_input import (
    Interpolation,
    Precision,
    ShaderInput,
)


def generate_varying_line(shader_input: ShaderInput, stage: ShaderStage):
    """
    Creates a text line from varying.def.sc for a specific shader input and stage.
    """
    line = ""
    is_instance_data = False

    if shader_input.interpolation != Interpolation.none:
        line += shader_input.interpolation.name + " "
    if shader_input.precision != Precision.none:
        line += shader_input.precision.name + " "
    line += shader_input.type.name + " "
    name = shader_input.name
    if name.startswith("instanceData"):
        name = f"i_data{int(name.removeprefix('instanceData'))+1}"
        is_instance_data = True
    elif stage == ShaderStage.Vertex:
        name = "a_" + shader_input.semantic.get_variable_name()
    else:
        name = "v_" + name
    line += f"{name} : {shader_input.semantic.get_name()};"

    return is_instance_data, line


def _postprocess_varying(code: str):
    """
    Formats generated varying.def code and corrects platform macros.
    """
    a_pattern = re.compile(r"^(.+? )(a_\w+)([\s]+: [\w]+;)", re.MULTILINE)
    i_pattern = re.compile(r"^(.+? )(i_\w+)([\s]+: [\w]+;)", re.MULTILINE)
    v_pattern = re.compile(r"^(.+? )(v_\w+)([\s]+: [\w]+;)", re.MULTILINE)

    for pattern in [a_pattern, i_pattern, v_pattern]:
        matches: list[re.Match] = re.findall(pattern, code)
        if not matches:
            continue
        max_type_len = max((len(x[0]) for x in matches))
        max_name_len = max((len(x[1]) for x in matches))

        def replacement(m: re.Match):
            g = m.groups()
            txt = g[0] + (max_type_len - len(g[0])) * " "
            txt += g[1] + (max_name_len - len(g[1])) * " "
            return txt + g[2]

        code = re.sub(pattern, replacement, code)

    for platform in ShaderPlatform:
        lang = "UNKNOWN"
        version = 1
        if platform.name.startswith("Direct3D_"):
            lang = "HLSL"
            if platform == ShaderPlatform.Direct3D_SM40:
                version = 400
            elif platform.name.startswith("Direct3D_SM"):
                version = 500
        elif platform.name.startswith("GLSL_") or platform.name.startswith("ESSL_"):
            lang = "GLSL"
            version = int(platform.name[-3:])
        elif platform == ShaderPlatform.Vulkan:
            lang = "SPIRV"
        elif platform == ShaderPlatform.Nvn:
            pass  # Don't know what should be used here.
        else:
            lang = platform.name.upper()

        lang = f"BGFX_SHADER_LANGUAGE_{lang}"
        macro = util.generate_flag_name_macro("platform", platform.name)
        code = (
            code.replace(f"defined({macro})", f"({lang} == {version})")
            .replace(f"#ifdef {macro}", f"#if {lang} == {version}")
            .replace(f"#ifndef {macro}", f"#if {lang} != {version}")
        )
    return code


def restore_varying(permutations: list[Variant], search_timeout: float = 10):
    _, code = restore_code(permutations, False, search_timeout=search_timeout)

    return _postprocess_varying(code)



from lazurite import util
from ..platform import ShaderPlatform
from ..stage import ShaderStage
from io import BytesIO
import struct, os


class BgfxUniform:
    name: str
    type_bits: int
    count: int
    reg_index: int
    reg_count: int

    def read(self, file: BytesIO):
        self.name = file.read(util.read_ubyte(file)).decode()
        self.type_bits = util.read_ubyte(file)
        self.count = util.read_ubyte(file)
        self.reg_index = util.read_ushort(file)
        self.reg_count = util.read_ushort(file)

        return self

    def write(self, file: BytesIO):
        util.write_ubyte(file, len(self.name))
        file.write(self.name.encode())
        util.write_ubyte(file, self.type_bits)
        util.write_ubyte(file, self.count)
        util.write_ushort(file, self.reg_index)
        util.write_ushort(file, self.reg_count)
        return self

    def serialize_properties(self):
        obj = {}
        obj["name"] = self.name
        obj["type_bits"] = self.type_bits
        obj["count"] = self.count
        obj["reg_index"] = self.reg_index
        obj["reg_count"] = self.reg_count
        return obj

    def load(self, object: dict):
        self.name = object["name"]
        self.type_bits = object["type_bits"]
        self.count = object["count"]
        self.reg_index = object["reg_index"]
        self.reg_count = object["reg_count"]
        return self


class BgfxShader:
    hash: int
    uniforms: list[BgfxUniform]
    group_size: list[int]
    shader_bytes: bytes
    attributes: list[int]  # Array of attribute IDs
    size: int

    def __init__(self) -> None:
        self.hash = 0
        self.uniforms = []
        self.group_size = []
        self.shader_bytes = b""
        self.attributes = []
        self.size = 0

    def read(self, file: BytesIO, platform: ShaderPlatform, stage: ShaderStage):
        header = file.read(3).decode()
        if not header in ["VSH", "FSH", "CSH"]:
            raise Exception(f'Unrecognized BGFX shader bin header "{header}"')

        version = util.read_ubyte(file)
        if not (version == 5 or version == 3 and header == "CSH"):
            raise Exception(f"Unsupported BGFX shader bin version: {version}")

        self.hash = util.read_ulong(file)
        self.uniforms = [
            BgfxUniform().read(file) for _ in range(util.read_ushort(file))
        ]

        if platform == ShaderPlatform.Metal and stage == ShaderStage.Compute:
            self.group_size = [
                util.read_ushort(file),
                util.read_ushort(file),
                util.read_ushort(file),
            ]
        else:
            self.group_size = []

        self.shader_bytes = file.read(util.read_ulong(file))
        util.read_ubyte(file)  # Padding (always 0)

        attribute_count = file.read(1)
        if len(attribute_count) != 0:
            self.attributes = [
                util.read_ushort(file)
                for _ in range(struct.unpack("<B", attribute_count)[0])
            ]
            self.size = util.read_ushort(file)
        else:
            self.attributes = []
            self.size = -1

        return self

    def write(self, file: BytesIO, platform: ShaderPlatform, stage: ShaderStage):
        bgfx_file = BytesIO()

        header = "FSH"
        version = 5
        if stage == ShaderStage.Vertex:
            header = "VSH"
        elif stage == ShaderStage.Compute:
            header = "CSH"
            version = 3

        bgfx_file.write(header.encode())
        util.write_ubyte(bgfx_file, version)

        util.write_ulong(bgfx_file, self.hash)

        util.write_ushort(bgfx_file, len(self.uniforms))
        for uniform in self.uniforms:
            uniform.write(bgfx_file)

        if platform == ShaderPlatform.Metal and stage == ShaderStage.Compute:
            for i in range(3):
                util.write_ushort(bgfx_file, self.group_size[i])

        util.write_ulong(bgfx_file, len(self.shader_bytes))
        bgfx_file.write(self.shader_bytes)

        util.write_ubyte(bgfx_file, 0)  # Padding

        if self.size != -1:
            util.write_ubyte(bgfx_file, len(self.attributes))
            for attribute in self.attributes:
                util.write_ushort(bgfx_file, attribute)

            util.write_ushort(bgfx_file, self.size)

        bgfx_file.seek(0)
        util.write_array(file, bgfx_file.read())
        return self

    def serialize_properties(self):
        obj = {}
        obj["hash"] = self.hash
        obj["uniforms"] = [uniform.serialize_properties() for uniform in self.uniforms]
        obj["group_size"] = self.group_size
        obj["attributes"] = self.attributes
        obj["size"] = self.size
        return obj

    def load(self, object: dict, path: str):
        bgfx_obj = object["bgfx_shader"]
        self.hash = bgfx_obj["hash"]
        self.uniforms = [
            BgfxUniform().load(uniform) for uniform in bgfx_obj["uniforms"]
        ]
        self.group_size = bgfx_obj["group_size"]
        self.attributes = bgfx_obj["attributes"]
        self.size = bgfx_obj["size"]

        with open(os.path.join(path, object["file_name"]), "rb") as f:
            self.shader_bytes = f.read()

        return self



from io import BytesIO
from enum import Enum

from lazurite import util
from ..precision import Precision


class Interpolation(Enum):
    none = None
    flat = 0
    smooth = 1
    noperspective = 2
    centroid = 3


class InputType(Enum):
    float = 0
    vec2 = 1
    vec3 = 2
    vec4 = 3
    int = 4
    ivec2 = 5
    ivec3 = 6
    ivec4 = 7
    uint = 8
    uvec2 = 9
    uvec3 = 10
    uvec4 = 11
    mat4 = 12


class InputSemantic:
    TYPES = [
        # TODO: potentially remove is_range_allowed and remove 0 from input names if they have it. (check if compiles first??)
        # (semantic, variable_name, is_range_allowed)
        ("POSITION", "position", False),
        ("NORMAL", "normal", False),
        ("TANGENT", "tangent", False),
        ("BITANGENT", "bitangent", False),
        ("COLOR", "color", True),  # range 0 - 3
        ("BLENDINDICES", "indices", False),
        ("BLENDWEIGHT", "weight", False),
        ("TEXCOORD", "texcoord", True),  # range 0 - 8
        ("UNKNOWN", "unknown", True),  # TODO: figure out (potentially VPOS?).
        ("FRONTFACING", "frontFacing", False),
    ]

    index: int
    sub_index: int

    def __init__(self, index=0, sub_index=0) -> None:
        self.index = index
        self.sub_index = sub_index

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, InputSemantic):
            return False

        return self.index == __value.index and self.sub_index == __value.sub_index

    def get_name(self) -> str:
        name, _, is_ranged = self.TYPES[self.index]
        return name + (str(self.sub_index) if is_ranged else "")

    def get_variable_name(self):
        _, name, is_ranged = self.TYPES[self.index]
        if is_ranged:
            name += str(self.sub_index)
        return name

    @classmethod
    def from_name(cls, name: str):
        semantic = cls()
        for i, (attr_type, _, _) in enumerate(cls.TYPES):
            if name.startswith(attr_type):
                semantic.index = i
                semantic.sub_index = int(
                    name.removeprefix(attr_type) or semantic.sub_index
                )
                break

        return semantic


class ShaderInput:
    name: str
    type: InputType
    semantic: InputSemantic
    per_instance: bool
    precision: Precision
    interpolation: Interpolation

    def __eq__(self, __value: object) -> bool:
        if type(__value) != ShaderInput:
            return False
        return (
            self.name == __value.name
            and self.type == __value.type
            and self.semantic == __value.semantic
            and self.per_instance == __value.per_instance
            and self.precision == __value.precision
            and self.interpolation == __value.interpolation
        )

    def read(self, file):
        self.name = util.read_string(file)
        self.type = InputType(util.read_ubyte(file))  # 0 - 4
        self.semantic = InputSemantic(util.read_ubyte(file), util.read_ubyte(file))
        self.per_instance = util.read_bool(file)

        if util.read_bool(file):
            self.precision = Precision(util.read_ubyte(file))
        else:
            self.precision = Precision.none

        if util.read_bool(file):
            self.interpolation = Interpolation(util.read_ubyte(file))  # 0-3
        else:
            self.interpolation = Interpolation.none

        return self

    def write(self, file: BytesIO):
        util.write_string(file, self.name)
        util.write_ubyte(file, self.type.value)
        util.write_ubyte(file, self.semantic.index)
        util.write_ubyte(file, self.semantic.sub_index)
        util.write_bool(file, self.per_instance)

        util.write_bool(file, self.precision != Precision.none)
        if self.precision != Precision.none:
            util.write_ubyte(file, self.precision.value)

        util.write_bool(file, self.interpolation != Interpolation.none)
        if self.interpolation != Interpolation.none:
            util.write_ubyte(file, self.interpolation.value)

        return self

    def serialize_properties(self):
        obj = {}
        obj["name"] = self.name
        obj["type"] = self.type.name
        obj["semantic"] = self.semantic.get_name()
        obj["per_instance"] = self.per_instance
        obj["precision"] = (
            self.precision.name if self.precision != Precision.none else ""
        )
        obj["interpolation"] = (
            self.interpolation.name if self.interpolation != Interpolation.none else ""
        )
        return obj

    def load(self, object: dict):
        self.name = object["name"]
        self.type = InputType[object["type"]]
        self.semantic = InputSemantic.from_name(object["semantic"])
        self.per_instance = object["per_instance"]
        self.precision = Precision[object["precision"] or "none"]
        self.interpolation = Interpolation[object["interpolation"] or "none"]
        return self



import pyjson5, copy, os
from .compiler_type import CompilerType
from lazurite.compiler.macro_define import MacroDefine
from .shader_file_overwrite import ShaderFileOverwrite
from lazurite.material.shader_pass.shader_definition import ShaderPlatform


class MaterialConfig:
    compiler_type: CompilerType
    compiler_options: list[str]
    macro_overwrite_pass: dict[str, list[MacroDefine]]
    macro_overwrite_flags: dict[str, dict[str, list[MacroDefine]]]
    file_overwrite_default: ShaderFileOverwrite
    file_overwrite_pass: dict[str, ShaderFileOverwrite]
    supported_platforms: set[ShaderPlatform]

    def __init__(self) -> None:
        self.compiler_type = CompilerType.SHADERC
        self.compiler_options = []
        self.macro_overwrite_pass = {}
        self.macro_overwrite_flags = {}
        self.file_overwrite_default = ShaderFileOverwrite()
        self.file_overwrite_pass = {}
        self.supported_platforms = set(ShaderPlatform)

    def read_from_json_file(self, file_path: str):
        if not os.path.isfile(file_path):
            return

        with open(file_path) as f:
            mat_json = pyjson5.load(f)
        self.read_json(mat_json)

    def read_json(self, json_data: dict):
        # Compiler config
        compiler_config: dict = json_data.get("compiler", {})

        self.compiler_type = (
            CompilerType.from_name(compiler_config.get("type", ""))
            or self.compiler_type
        )
        self.compiler_options = compiler_config.get("options", self.compiler_options)

        # Macro overwrite
        macro_overwrite: dict = json_data.get("macro_overwrite", {})
        flags: dict[str, dict[str, list[str]]] = macro_overwrite.get("flags", {})
        for _, values in flags.items():
            for key, macros in values.items():
                if type(macros) == str:
                    macros = [macros]
                values[key] = [MacroDefine.from_string(x) for x in macros]
        self.macro_overwrite_flags = flags

        passes: dict = macro_overwrite.get("passes", {})
        for key, value in passes.items():
            if type(value) == str:
                value = [value]
            passes[key] = [MacroDefine.from_string(m) for m in value]

        self.macro_overwrite_pass = passes

        # File overwrite
        file_overwrite: dict = json_data.get("file_overwrite", {})
        self.file_overwrite_default.read_json(file_overwrite.get("default", {}))
        if "passes" in file_overwrite:
            self.file_overwrite_pass = {}
        for key, value in file_overwrite.get("passes", {}).items():
            file_overwrite_pass = copy.copy(self.file_overwrite_default)
            file_overwrite_pass.read_json(value)
            self.file_overwrite_pass[key] = file_overwrite_pass

        # Platforms
        if "supported_platforms" in json_data:
            self.supported_platforms = {
                ShaderPlatform[x] for x in json_data["supported_platforms"]
            }



import subprocess
from lazurite.material.stage import ShaderStage
from lazurite.material.platform import ShaderPlatform
from lazurite.tempfile import CustomTempFile

from .macro_define import MacroDefine


class DxcCompiler:
    dxc_path: str

    def __init__(self, dxc_paths: list[str] | str = None) -> None:
        if dxc_paths is None:
            dxc_paths = ["dxc", "./dxc"]
        elif isinstance(dxc_paths, str):
            dxc_paths = [dxc_paths]

        self.dxc_path = ""

        for path in dxc_paths:
            try:
                result = subprocess.run([path, "-help"], capture_output=True)
            except FileNotFoundError:
                pass
            else:
                self.dxc_path = path
                break

        if not self.dxc_path:
            raise Exception(
                f"Error! No valid DXC compiler was found in the list {dxc_paths}"
            )

        if result.returncode:
            print(result.stdout.decode())
            print(result.stderr.decode())
            result.check_returncode()

    def compile(
        self,
        file: str,
        platform: ShaderPlatform = ShaderPlatform.Direct3D_SM65,
        stage: ShaderStage = ShaderStage.Compute,
        entry_point: str = None,
        defines: list[MacroDefine] = None,
        options: list[str] = None,
    ):
        if not platform.name.startswith("Direct3D_SM"):
            raise Exception(
                f"{platform.name} shaders cannot be compiled with DXC compiler!"
            )

        args = [self.dxc_path, file]

        if stage == ShaderStage.Compute:
            profile = "cs"  # Compute Shader
        elif stage == ShaderStage.Vertex:
            profile = "vs"  # Vertex Shader
        else:
            profile = "ps"  # Pixel Shader
        version = platform.name.removeprefix("Direct3D_SM")
        args.extend(("-T", "_".join((profile, *version))))

        if entry_point:
            args.extend(("-E", entry_point))

        if defines:
            for define in defines:
                args.extend(("-D", define.format_dxc()))

        if options:
            args.extend(options)

        with CustomTempFile() as f:
            f.close()
            args.extend(("-Fo", f.name))

            result = subprocess.run(args)

            if result.stderr:
                print(result.stderr.decode())
            if result.stdout:
                print(result.stdout.decode())

            result.check_returncode()

            with open(f.name, "rb") as f:
                return f.read()



import json, os
from io import BytesIO
from enum import Enum

from lazurite import util
from .precision import Precision


class BufferAccess(Enum):
    undefined = 0
    readonly = 1
    writeonly = 2
    readwrite = 3


class BufferType(Enum):
    texture2D = 0
    texture2DArray = 1
    external2D = 2  # ?
    texture3D = 3  # ?
    textureCube = 4
    structBuffer = 5
    rawBuffer = 6
    accelerationStructure = 7  # ?
    shadow2D = 8
    shadow2DArray = 9

    # ? -> never observed in-game


class Buffer:
    class CustomTypeInfo:
        struct: str
        size: int

        def __init__(self, name="", size=0):
            self.struct = name
            self.size = size

    name: str
    reg1: int
    access: BufferAccess
    precision: Precision
    unordered_access: bool
    type: BufferType
    texture_format: str
    always_one: int
    reg2: int
    default_texture: str
    unknown_string: str
    custom_type_info: CustomTypeInfo | None

    def __init__(self):
        self.name = ""
        self.reg1 = 0
        self.access = BufferAccess.readonly
        self.precision = Precision.lowp
        self.unordered_access = False
        self.type = BufferType.texture2D
        self.texture_format = ""
        self.always_one = 1
        self.reg2 = 0
        self.default_texture = ""
        self.unknown_string = ""
        self.custom_type_info = None

    def read(self, file: BytesIO):
        self.name = util.read_string(file)
        self.reg1 = util.read_ushort(file)
        self.access = BufferAccess(util.read_ubyte(file))  # 1 2 3
        self.precision = Precision(util.read_ubyte(file))  # 0 2
        self.unordered_access = util.read_bool(file)
        self.type = BufferType(util.read_ubyte(file))  # 0 - 9
        # Values according to bgfx_compute.sh: (empty string) r32ui rg32ui rgba32ui r32f r16f rg16f rgba16f rgba8 rg8 r8 rgba32f
        self.texture_format = util.read_string(file)
        self.always_one = util.read_ulong(file)  # 1
        self.reg2 = util.read_ubyte(file)  # same as reg1

        if util.read_bool(file):
            self.default_texture = util.read_string(file)  # white
        else:
            self.default_texture = ""

        if util.read_bool(file):
            self.unknown_string = util.read_string(file)
            # print(self.unknown_string)  # TODO: remove

        if util.read_bool(file):  # CustomTypeInfo
            self.custom_type_info = self.CustomTypeInfo()
            self.custom_type_info.struct = util.read_string(file)
            self.custom_type_info.size = util.read_ulong(file)

        return self

    def write(self, file: BytesIO):
        util.write_string(file, self.name)
        util.write_ushort(file, self.reg1)
        util.write_ubyte(file, self.access.value)
        util.write_ubyte(file, self.precision.value)
        util.write_bool(file, self.unordered_access)
        util.write_ubyte(file, self.type.value)
        util.write_string(file, self.texture_format)
        util.write_ulong(file, self.always_one)
        util.write_ubyte(file, self.reg2)

        util.write_bool(file, self.default_texture != "")
        if self.default_texture != "":
            util.write_string(file, self.default_texture)

        util.write_bool(file, self.unknown_string != "")
        if self.unknown_string != "":
            util.write_string(file, self.unknown_string)

        util.write_bool(file, self.custom_type_info != None)
        if self.custom_type_info != None:
            util.write_string(file, self.custom_type_info.struct)
            util.write_ulong(file, self.custom_type_info.size)

        return self

    def serialize_properties(self):
        obj = {
            "name": self.name,
            "reg1": self.reg1,
            "reg2": self.reg2,
            "type": self.type.name,
            "precision": self.precision.name,
            "access": self.access.name,
            "texture_format": self.texture_format,
            "default_texture": self.default_texture,
            "unordered_access": self.unordered_access,
            "always_one": self.always_one,
            "unknown_string": self.unknown_string,
            "custom_type_info": {},
        }
        if self.custom_type_info != None:
            obj["custom_type_info"]["struct"] = self.custom_type_info.struct
            obj["custom_type_info"]["size"] = self.custom_type_info.size

        return obj

    def store(self, path: str = "."):
        with open(os.path.join(path, f"{self.name}.json"), "w") as f:
            json.dump(self.serialize_properties(), f, indent=4)

        return self

    def load(self, object: dict, path: str):
        self.name = object.get("name", self.name)
        self.reg1 = object.get("reg1", self.reg1)
        self.access = BufferAccess[object.get("access", self.access.name)]
        self.precision = Precision[object.get("precision", self.precision.name)]
        self.unordered_access = object.get("unordered_access", self.unordered_access)
        self.type = BufferType[object.get("type", self.type.name)]
        self.texture_format = object.get("texture_format", self.texture_format)
        self.always_one = object.get("always_one", self.always_one)
        self.unknown_string = object.get("unknown_string", self.unknown_string)
        self.reg2 = object.get("reg2", self.reg2)
        self.default_texture = object.get("default_texture", self.default_texture)

        if "custom_type_info" in object:
            if len(object["custom_type_info"]) > 0:
                if self.custom_type_info == None:
                    self.custom_type_info = self.CustomTypeInfo()

                self.custom_type_info.struct = object["custom_type_info"].get(
                    "struct", self.custom_type_info.struct
                )
                self.custom_type_info.size = object["custom_type_info"].get(
                    "size", self.custom_type_info.size
                )

            else:
                self.custom_type_info = None

        return self



from io import BytesIO
import os, json

from lazurite import util
from .variant import Variant
from ..platform import ShaderPlatform
from ..stage import ShaderStage
from .blend_mode import BlendMode
from .supported_platforms import SupportedPlatforms


class Pass:
    name: str
    supported_platforms: SupportedPlatforms
    fallback_pass: str
    default_blend_mode: BlendMode
    default_variant: dict[str, str]
    variants: list[Variant]

    def __init__(self):
        self.name = ""
        self.supported_platforms = SupportedPlatforms()
        self.fallback_pass = ""
        self.default_blend_mode = BlendMode.Unspecified
        self.default_variant = {}
        self.variants = []

    def read(self, file: BytesIO):
        self.name = util.read_string(file)
        self.supported_platforms = SupportedPlatforms(util.read_string(file))
        self.fallback_pass = util.read_string(
            file
        )  # (empty string) Fallback DoCheckerboarding DepthOnlyFallback

        if util.read_bool(file):  # Has default blend mode
            self.default_blend_mode = BlendMode[util.read_ushort(file)]

        self.default_variant = {}
        default_flag_count = util.read_ushort(file)
        for _ in range(default_flag_count):
            key = util.read_string(file)
            self.default_variant[key] = util.read_string(file)

        self.variants = [Variant().read(file) for _ in range(util.read_ushort(file))]

        return self

    def write(self, file: BytesIO):
        util.write_string(file, self.name)
        util.write_string(file, self.supported_platforms.get_bit_string())
        util.write_string(file, self.fallback_pass)

        util.write_bool(file, self.default_blend_mode != BlendMode.Unspecified)
        if self.default_blend_mode != BlendMode.Unspecified:
            util.write_ushort(file, self.default_blend_mode.value)

        util.write_ushort(file, len(self.default_variant))
        for key in self.default_variant:
            util.write_string(file, key)
            util.write_string(file, self.default_variant[key])

        util.write_ushort(file, len(self.variants))
        for variant in self.variants:
            variant.write(file)
        return self

    def serialize_properties(self):
        obj = {}
        obj["name"] = self.name
        obj["supported_platforms"] = self.supported_platforms.serialize()
        obj["fallback_pass"] = self.fallback_pass
        obj["default_blend_mode"] = (
            self.default_blend_mode.name
            if self.default_blend_mode != BlendMode.Unspecified
            else ""
        )
        obj["default_variant"] = self.default_variant
        obj["variants"] = []

        for i in range(len(self.variants)):
            obj["variants"].append(self.variants[i].serialize_properties(i))

        return obj

    def store(self, path: str = ".", skip_shaders=False):
        pass_dir = os.path.join(path, self.name)

        with open(os.path.join(path, f"{self.name}.json"), "w") as f:
            json.dump(self.serialize_properties(), f, indent=4)

        if skip_shaders:
            return self

        os.mkdir(pass_dir)

        for i in range(len(self.variants)):
            for shader in self.variants[i].shaders:
                with open(
                    os.path.join(pass_dir, shader.get_shader_file_name(i)), "wb"
                ) as f:
                    f.write(shader.bgfx_shader.shader_bytes)

        return self

    def load(self, object: dict, path: str):
        self.name = object.get("name", self.name)
        self.supported_platforms.load(object.get("supported_platforms", {}))
        self.fallback_pass = object.get("fallback_pass", self.fallback_pass)
        mode = object.get("default_blend_mode", None)
        if mode != None:
            self.default_blend_mode = BlendMode(mode) if mode else BlendMode.Unspecified
        self.default_variant = object.get("default_variant", self.default_variant)

        if "variants" in object:
            self.variants = [
                Variant().load(variant, os.path.join(path, self.name))
                for variant in object["variants"]
            ]
        return self

    def label(self, material_name: str):
        for variant_index, variant in enumerate(self.variants):
            variant.label(material_name, self.name, variant_index)

        return self

    def sort_variants(self):
        self.default_variant = dict(sorted(self.default_variant.items()))

        for variant in self.variants:
            variant.flags = dict(sorted(variant.flags.items()))

        self.variants.sort(key=lambda x: str(x.flags))

    def get_platforms(self):
        platforms: set[ShaderPlatform] = set()
        for variant in self.variants:
            platforms.update(variant.get_platforms())

        return platforms

    def get_stages(self):
        stages: set[ShaderStage] = set()
        for variant in self.variants:
            stages.update(variant.get_stages())

        return stages

    def merge_variants(self, other: "Pass"):
        for other_variant in other.variants:
            matching_variant = next(
                (v for v in self.variants if v.flags == other_variant.flags), None
            )
            if matching_variant is None:
                self.variants.append(other_variant)
            else:
                matching_variant.merge_variant(other_variant)

    def get_flag_definitions(self):
        """
        Returns a dict of all possible flag keys and their values.
        """
        definitions = {key: {value} for key, value in self.default_variant.items()}

        for variant in self.variants:
            for key, value in variant.flags.items():
                if key not in definitions:
                    definitions[key] = set()
                definitions[key].add(value)
        return definitions

    def add_platforms(self, platforms: set[ShaderPlatform]):
        for variant in self.variants:
            variant.add_platforms(platforms)

    def remove_platforms(self, platforms: set[ShaderPlatform]):
        for variant in self.variants:
            variant.remove_platforms(platforms)



import enum


class CompilerType(enum.Enum):
    SHADERC = enum.auto()
    DXC = enum.auto()

    @classmethod
    def from_name(cls, name: str):
        if name == "shaderc":
            return cls.SHADERC
        if name == "dxc":
            return cls.DXC



import copy
from io import BytesIO

from lazurite import util
from .shader_definition import ShaderDefinition
from ..platform import ShaderPlatform
from ..stage import ShaderStage
from .bgfx_shader import BgfxShader


class Variant:
    is_supported: bool
    flags: dict[str, str]
    shaders: list[ShaderDefinition]

    def __init__(self) -> None:
        self.is_supported = False
        self.flags = {}
        self.shaders = []

    def read(self, file: BytesIO):
        self.is_supported = util.read_bool(file)
        flag_count = util.read_ushort(file)
        shader_count = util.read_ushort(file)

        self.flags = {}
        for _ in range(flag_count):
            key = util.read_string(file)
            self.flags[key] = util.read_string(file)

        self.shaders = [ShaderDefinition().read(file) for _ in range(shader_count)]

        return self

    def write(self, file: BytesIO):
        util.write_bool(file, self.is_supported)
        util.write_ushort(file, len(self.flags))
        util.write_ushort(file, len(self.shaders))

        for key in self.flags:
            util.write_string(file, key)
            util.write_string(file, self.flags[key])

        for shader in self.shaders:
            shader.write(file)

        return self

    def serialize_properties(self, index: int):
        obj = {}
        obj["is_supported"] = self.is_supported
        obj["flags"] = self.flags
        obj["shaders"] = [shader.serialize_properties(index) for shader in self.shaders]

        return obj

    def load(self, object: dict, path: str):
        self.is_supported = object.get("is_supported", self.is_supported)
        self.flags = object.get("flags", self.flags)

        if "shaders" in object:
            self.shaders = [
                ShaderDefinition().load(shader, path) for shader in object["shaders"]
            ]
        return self

    def label(self, material_name: str, pass_name: str, variant_index: int):
        for shader in self.shaders:
            shader.label(
                material_name, pass_name, variant_index, self.is_supported, self.flags
            )

        return self

    def get_platforms(self):
        platforms: set[ShaderPlatform] = set()
        for shader in self.shaders:
            platforms.add(shader.platform)

        return platforms

    def get_stages(self):
        stages: set[ShaderStage] = set()
        for shader in self.shaders:
            stages.add(shader.stage)

        return stages

    def merge_variant(self, other: "Variant"):
        for other_shader in other.shaders:
            matching_shader = next(
                (
                    v
                    for v in self.shaders
                    if v.platform == other_shader.platform
                    and v.stage == other_shader.stage
                ),
                None,
            )
            if matching_shader is None:
                self.shaders.append(other_shader)

    def add_platforms(self, platforms: set[ShaderPlatform]):
        current_platforms = self.get_platforms()
        missing_platforms = platforms.difference(current_platforms)
        stages = self.get_stages()

        for platform in missing_platforms:
            for stage in stages:
                shader = next(
                    (x for x in self.shaders if x.stage == stage), ShaderDefinition()
                )
                shader = copy.deepcopy(shader)
                shader.stage = stage
                shader.platform = platform
                shader.bgfx_shader = BgfxShader()
                self.shaders.append(shader)

    def remove_platforms(self, platforms: set[ShaderPlatform]):
        self.shaders = [s for s in self.shaders if s.platform not in platforms]



from enum import Enum


class ShaderPlatform(Enum):
    Direct3D_SM40 = 0
    Direct3D_SM50 = 1
    Direct3D_SM60 = 2
    Direct3D_SM65 = 3
    Direct3D_XB1 = 4
    Direct3D_XBX = 5
    GLSL_120 = 6
    GLSL_430 = 7
    ESSL_100 = 8
    ESSL_300 = 9
    ESSL_310 = 10
    Metal = 11
    Vulkan = 12
    Nvn = 13
    PSSL = 14
    Unknown = 15

    def file_extension(self):
        if self.name.startswith("Direct3D"):
            return "dxbc"

        elif self.name.startswith("GLSL") or self.name.startswith("ESSL"):
            return "glsl"

        elif self.name == "Metal":
            return "metal"

        elif self.name == "Vulkan":
            return "spirv"

        else:
            return "bin"



import contextlib, tempfile, os

# from typing import TypeVar, Callable, Any, Generic

# F = TypeVar("F", bound=Callable[..., Any])


# class copy_signature(Generic[F]):
#     def __init__(self, target: F) -> None: ...
#     def __call__(self, wrapped: Callable[..., Any]) -> F: ...


# this is a backport of delete_on_close functionality from 3.12
@contextlib.contextmanager
# @copy_signature(tempfile.NamedTemporaryFile)
def CustomTempFile(*args, **kwargs):
    """
    Context manager for tempfile.NamedTemporaryFile which implements automatic file deletion on exit.
    """
    kwargs["delete"] = False
    file = tempfile.NamedTemporaryFile(*args, **kwargs)
    try:
        yield file
    finally:
        file.close()
        with contextlib.suppress(FileNotFoundError):
            os.remove(file.name)



import os, pyjson5, pcpp, pathlib
from concurrent.futures import ThreadPoolExecutor

from .material_config import MaterialConfig
from .compiler_type import CompilerType
from .project_config import ProjectConfig

from lazurite import util
from lazurite.material import Material
from lazurite.material.shader_pass.shader_definition import ShaderPlatform
from lazurite.material.stage import ShaderStage
from lazurite.compiler.shaderc import ShadercCompiler, generate_bgfx_defines
from lazurite.compiler.dxc import DxcCompiler
from lazurite.compiler.macro_define import MacroDefine
from lazurite.material.shader_pass import Pass
from lazurite.material.shader_pass.variant import Variant
from lazurite.material.shader_pass.shader_definition import ShaderDefinition
from lazurite.tempfile import CustomTempFile


def _merge_source_by_name(
    name: str,
    platforms: list[ShaderPlatform],
    merge_source: list[str],
    cache: dict[str, tuple[str, set[ShaderPlatform]]],
):
    mat = None
    for platform in platforms:
        for path in merge_source:
            found_platform = False
            if path in cache:
                cache_name, cache_platforms = cache[path]
                if cache_name == name and platform in cache_platforms:
                    found_platform = True
                    temp_mat = Material.load_bin_file(path)
                    if mat is None:
                        mat = temp_mat
                    else:
                        mat.merge_variants(temp_mat)
            else:
                temp_mat = Material.load_bin_file(path)
                mat_platforms = temp_mat.get_platforms()
                if temp_mat.name == name and platform in mat_platforms:
                    found_platform = True
                    if mat is None:
                        mat = temp_mat
                    else:
                        mat.merge_variants(temp_mat)
                cache[path] = (temp_mat.name, mat_platforms)
            if found_platform:
                break
    return mat


def _merge_source_by_path(
    name: str,
    platforms: list[ShaderPlatform],
    merge_source: list[str],
    cache: dict[str, tuple[str, set[ShaderPlatform]]],
):
    mat = None
    for platform in platforms:
        for path in merge_source:
            if not path.endswith(name):
                continue
            found_platform = False
            if path in cache:
                if platform in cache[path][1]:
                    found_platform = True
                    temp_mat = Material.load_bin_file(path)
                    if mat is None:
                        mat = temp_mat
                    else:
                        mat.merge_variants(temp_mat)
            else:
                temp_mat = Material.load_bin_file(path)
                mat_platforms = temp_mat.get_platforms()
                if platform in mat_platforms:
                    found_platform = True
                    if mat is None:
                        mat = temp_mat
                    else:
                        mat.merge_variants(temp_mat)
                cache[path] = (temp_mat.name, mat_platforms)
            if found_platform:
                break
    return mat


def _merge_source_materials(
    mat_dir: os.DirEntry,
    platforms: list[ShaderPlatform],
    merge_source: list[str],
    material_cache: dict[str, tuple[str, set[ShaderPlatform]]],
):
    name = mat_dir.name + Material.EXTENSION
    use_name = False
    path = os.path.join(mat_dir, "material.json")
    if os.path.isfile(path):
        with open(path) as f:
            json_data = pyjson5.load(f)
        if "name" in json_data:
            name = json_data["name"]
            use_name = True

    if use_name:
        return _merge_source_by_name(name, platforms, merge_source, material_cache)
    else:
        return _merge_source_by_path(name, platforms, merge_source, material_cache)


def _generate_defines(
    config: ProjectConfig,
    material: Material,
    mat_config: MaterialConfig,
    shader_pass: Pass,
    variant: Variant,
):
    defines = config.macros[:]
    if mat_config.compiler_type is CompilerType.SHADERC:
        defines.append(MacroDefine("BGFX_CONFIG_MAX_BONES", 4))

    defines.extend(MacroDefine(f"s_{s.name}_REG", s.reg1) for s in material.buffers)

    if shader_pass.name in mat_config.macro_overwrite_pass:
        defines.extend(mat_config.macro_overwrite_pass[shader_pass.name])
    else:
        defines.append(MacroDefine(util.generate_pass_name_macro(shader_pass.name)))

    for key, value in variant.flags.items():
        if (
            key in mat_config.macro_overwrite_flags
            and value in mat_config.macro_overwrite_flags[key]
        ):
            defines.extend(mat_config.macro_overwrite_flags[key][value])
        else:
            defines.append(MacroDefine(util.generate_flag_name_macro(key, value)))

    return defines


def _compile_bgfx_shader_async(
    shaderc_compiler: ShadercCompiler,
    code_path: str,
    stage: ShaderStage,
    platform: ShaderPlatform,
    varying_path: str,
    defines: list[MacroDefine],
    options: list[str],
):
    if stage == ShaderStage.Compute:
        compiled_shader = shaderc_compiler.compile(
            code_path, platform, stage, None, None, defines, options
        )
    else:
        parser = pcpp.Preprocessor()
        parser.line_directive = None

        varying_defines = defines + generate_bgfx_defines(platform, stage)
        for macro in varying_defines:
            parser.define(macro.format_cpp())

        with open(varying_path) as f:
            parser.parse(f)

        with CustomTempFile("w+") as f:
            parser.write(f)
            f.close()
            # f.flush()

            compiled_shader = shaderc_compiler.compile(
                code_path, platform, stage, f.name, None, defines, options
            )

    return compiled_shader


def _get_material_folders(proj_config: ProjectConfig, project_path: str):
    material_folders: set[pathlib.Path] = set()
    proj_path = pathlib.Path(project_path)
    for pattern in proj_config.include_patterns:
        for path in proj_path.glob(pattern):
            if path.is_dir() and not any(
                path.match(p) for p in proj_config.exclude_patterns
            ):
                material_folders.add(path)
    return sorted(list(material_folders))


def compile(
    project_path: str,
    profiles: list[str],
    output_folder: str = "",
    defines: list[MacroDefine] = None,
    shaderc_path: str = None,
    dxc_path: str = None,
    shaderc_args: list[str] = None,
    dxc_args: list[str] = None,
    max_workers: int = None,
):
    if not os.path.isdir(project_path):
        raise Exception(f'Failed to compile project: "{project_path}" is not a folder.')

    if profiles is None:
        profiles = []

    if shaderc_args is None:
        shaderc_args = []

    if dxc_args is None:
        dxc_args = []

    proj_config = ProjectConfig()
    proj_config.read_json_file(os.path.join(project_path, "project.json"), profiles)
    platforms_set = set(proj_config.platforms)

    shaderc_compiler: ShadercCompiler = None
    dxc_compiler: DxcCompiler = None

    # {path: (name, {platforms})}
    material_cache: dict[str, tuple[str, set[ShaderPlatform]]] = {}

    for mat_dir in _get_material_folders(proj_config, project_path):
        shaders: list[ShaderDefinition] = []
        arg_defines: list[list[MacroDefine]] = []
        arg_code_path: list[str] = []
        arg_stage: list[ShaderStage] = []
        arg_platform: list[ShaderPlatform] = []
        arg_varying: list[str] = []
        arg_entry_point: list[str] = []

        print(mat_dir.name)

        mat_config = MaterialConfig()
        mat_config.read_from_json_file(os.path.join(mat_dir, "config.json"))

        # Target platforms AND platforms supported by material.
        compilable_platforms = platforms_set.intersection(
            mat_config.supported_platforms
        )

        # Load and merge vanilla target platform materials.
        material = (
            _merge_source_materials(
                mat_dir, proj_config.platforms, proj_config.merge_source, material_cache
            )
            or Material()
        )

        # If it failed to load any source materials.
        if not material.get_platforms():
            # Load and merge vanilla materials with any platforms.
            material = (
                _merge_source_materials(
                    mat_dir,
                    set(ShaderPlatform),
                    proj_config.merge_source,
                    material_cache,
                )
                or material
            )
            # Add necessary platforms.
            material.add_platforms(compilable_platforms)
            # Remove other unwanted platforms.
            material.remove_platforms(
                set(ShaderPlatform).difference(compilable_platforms)
            )
        else:
            material.add_platforms(compilable_platforms)

        material.load_unpacked_material(mat_dir)

        # Gather shaders for compilation.
        for shader_pass in material.passes:
            for variant in shader_pass.variants:
                for shader in variant.shaders:
                    if shader.platform not in compilable_platforms:
                        continue

                    file_overwrite = mat_config.file_overwrite_pass.get(
                        shader_pass.name, mat_config.file_overwrite_default
                    )
                    code_path = os.path.normpath(
                        os.path.join(mat_dir, file_overwrite.get_stage(shader.stage))
                    )
                    if not os.path.isfile(code_path):
                        continue

                    if mat_config.compiler_type is CompilerType.SHADERC:
                        varying_path = None
                        if shaderc_compiler is None:
                            shaderc_compiler = ShadercCompiler(shaderc_path)
                        if shader.stage is not ShaderStage.Compute:
                            varying_path = os.path.normpath(
                                os.path.join(mat_dir, file_overwrite.varying)
                            )
                            if not os.path.isfile(varying_path):
                                print(
                                    f'Varying file "{varying_path}" was not found! Skipping current shader.'
                                )
                                continue
                        arg_varying.append(varying_path)
                    else:
                        if dxc_compiler is None:
                            dxc_compiler = DxcCompiler(dxc_path)
                        arg_entry_point.append(
                            file_overwrite.entry_point or shader_pass.name
                        )

                    shaders.append(shader)
                    arg_code_path.append(code_path)
                    arg_stage.append(shader.stage)
                    arg_platform.append(shader.platform)
                    arg_defines.append(
                        _generate_defines(
                            proj_config, material, mat_config, shader_pass, variant
                        )
                        + defines
                    )

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            if mat_config.compiler_type is CompilerType.SHADERC:
                results = executor.map(
                    _compile_bgfx_shader_async,
                    len(shaders) * [shaderc_compiler],
                    arg_code_path,
                    arg_stage,
                    arg_platform,
                    arg_varying,
                    arg_defines,
                    len(shaders) * [mat_config.compiler_options + dxc_args],
                )
            else:
                results = executor.map(
                    dxc_compiler.compile,
                    arg_code_path,
                    arg_platform,
                    arg_stage,
                    arg_entry_point,
                    arg_defines,
                    len(shaders) * [mat_config.compiler_options + shaderc_args],
                )

        for shader, result in zip(shaders, results):
            if mat_config.compiler_type is CompilerType.SHADERC:
                shader.bgfx_shader = result
            else:
                shader.bgfx_shader.shader_bytes = result

        # Filter empty shaders.
        for shader_pass in material.passes:
            for variant in shader_pass.variants:
                new_shaders = []
                for shader in variant.shaders:
                    if shader.bgfx_shader.shader_bytes:
                        new_shaders.append(shader)
                variant.shaders = new_shaders

        filepath = output_folder or os.path.dirname(project_path)
        filepath = os.path.join(filepath, mat_dir.name + Material.EXTENSION)
        with open(filepath, "wb") as f:
            material.write(f)



from ..platform import ShaderPlatform


class SupportedPlatforms:
    platforms: dict[ShaderPlatform, bool]

    def _validate_bit_string(self, platforms_bit_string: str) -> str:
        if any(c not in "01" for c in platforms_bit_string):
            print(
                f"Warning: Invalid supported platforms bit field {platforms_bit_string}"
            )
            return "1" * len(ShaderPlatform)
        return platforms_bit_string

    def _format_bit_string(self, platforms_bit_string: str):
        # Truncate everything after 16 symbols.
        platforms_bit_string = platforms_bit_string[: len(ShaderPlatform)]

        # In case there are less than 16, add leading zeros.
        platforms_bit_string = platforms_bit_string.zfill(len(ShaderPlatform))

        return platforms_bit_string

    def __init__(self, platforms_bit_string: str = "1" * len(ShaderPlatform)) -> None:
        platforms_bit_string = self._validate_bit_string(platforms_bit_string)
        platforms_bit_string = self._format_bit_string(platforms_bit_string)

        self.platforms = {
            platform: bit == "1"
            for platform, bit in zip(ShaderPlatform, platforms_bit_string)
        }

    def get_bit_string(self) -> str:
        return "".join("1" if self.platforms[p] else "0" for p in ShaderPlatform)

    def load(self, json_data: dict[str, bool]):
        for key, value in json_data.items():
            self.platforms[ShaderPlatform[key]] = value

    def serialize(self):
        return {p.name: v for p, v in self.platforms.items()}



import struct, json, os
from enum import Enum
from io import BytesIO

from lazurite import util


class UniformType(Enum):
    vec4 = 2
    mat3 = 3
    mat4 = 4
    external = 5


class Uniform:

    name: str
    type: UniformType
    count: int
    default: list[float]

    def __init__(self):
        self.name = ""
        self.type = UniformType.vec4
        self.count = 0
        self.default = []

    def read(self, file: BytesIO):
        self.name = util.read_string(file)
        self.type = UniformType(util.read_ushort(file))

        self.default = []
        if 2 <= self.type.value <= 4:
            self.count = util.read_ulong(file)
            hasData = util.read_bool(file)

        if self.type.value == 2:  # vec4
            if hasData:
                self.default = struct.unpack("<" + 4 * "f", file.read(4 * 4))

        elif self.type.value == 3:  # mat3
            if hasData:
                self.default = struct.unpack("<" + 9 * "f", file.read(4 * 9))

        elif self.type.value == 4:  # mat4
            if hasData:
                self.default = struct.unpack("<" + 16 * "f", file.read(4 * 16))

        elif self.type.value == 5:  # external
            pass

        else:
            raise Exception(f'Urecognized type "{self.type}"')

        return self

    def write(self, file: BytesIO):
        util.write_string(file, self.name)
        util.write_ushort(file, self.type.value)

        if 2 <= self.type.value <= 4:
            util.write_ulong(file, self.count)
            util.write_bool(file, len(self.default) > 0)

        if len(self.default) > 0:
            file.write(struct.pack("<" + "f" * len(self.default), *self.default))

        return self

    def serialize_properties(self):
        obj = {}
        obj["name"] = self.name
        obj["type"] = self.type.name
        obj["count"] = self.count
        obj["default"] = self.default
        return obj

    def store(self, path: str = "."):
        with open(os.path.join(path, f"{self.name}.json"), "w") as f:
            json.dump(self.serialize_properties(), f, indent=4)

        return self

    def load(self, object: dict, path: str):
        self.name = object.get("name", self.name)
        self.type = UniformType[object.get("type", self.type.name)]
        self.count = object.get("count", self.count)
        self.default = object.get("default", self.default)
        return self



import subprocess

from lazurite.material.stage import ShaderStage
from lazurite.material.platform import ShaderPlatform
from lazurite.material.shader_pass.bgfx_shader import BgfxShader
from lazurite.tempfile import CustomTempFile

from .macro_define import MacroDefine


class ShadercCompiler:
    shaderc_path: str

    def __init__(self, shaderc_paths: list[str] | str = None) -> None:
        if shaderc_paths is None:
            shaderc_paths = ["shaderc", "./shaderc"]
        elif isinstance(shaderc_paths, str):
            shaderc_paths = [shaderc_paths]

        self.shaderc_path = ""

        for path in shaderc_paths:
            try:
                result = subprocess.run([path, "-v"], capture_output=True)
            except FileNotFoundError:
                pass
            else:
                self.shaderc_path = path
                break

        if not self.shaderc_path:
            raise Exception(
                f"Error! No valid SHADERC compiler was found in the list {shaderc_paths}"
            )

        if result.returncode:
            print(result.stdout.decode())
            print(result.stderr.decode())
            result.check_returncode()

    def compile(
        self,
        file: str,
        platform: ShaderPlatform = ShaderPlatform.ESSL_310,
        stage: ShaderStage = ShaderStage.Fragment,
        varying_def: str = None,
        include: list[str] = None,
        defines: list[MacroDefine] = None,
        options: list[str] = None,
    ) -> BgfxShader:
        args = [self.shaderc_path]

        args.extend(("-f", file))

        device = ""
        if platform == ShaderPlatform.Metal:
            device = "ios"
        elif platform.name.startswith("ESSL"):
            device = "android"
        elif platform.name.startswith("GLSL"):
            device = "linux"
        elif platform.name.startswith("Direct3D"):
            device = "windows"
        if device:
            args.extend(("--platform", device))

        profile = ""
        if platform == ShaderPlatform.Direct3D_SM40:
            profile = "s_4_0"
        elif platform == ShaderPlatform.Vulkan:
            profile = "spirv"
        elif platform in {ShaderPlatform.Metal, ShaderPlatform.PSSL}:
            profile = platform.name.lower()
        elif platform.name.startswith("Direct3D_SM"):
            profile = "s_5_0"
        elif platform.name.startswith("GLSL"):
            profile = platform.name.removeprefix("GLSL_")
        elif platform.name.startswith("ESSL"):
            profile = platform.name.removeprefix("ESSL_") + "_es"
        if profile:
            args.extend(("-p", profile))

        args.append("--type")
        if stage == ShaderStage.Compute:
            args.append("compute")
        elif stage == ShaderStage.Vertex:
            args.append("vertex")
        else:
            args.append("fragment")

        if varying_def is not None:
            args.extend(("--varyingdef", varying_def))

        if include:
            for item in include:
                args.extend(("-i", item))

        if defines:
            args.extend(("--define", ";".join(d.format_bgfx() for d in defines)))

        if options:
            args.extend(options)

        with CustomTempFile() as f:
            args.extend(("-o", f.name))

            result = subprocess.run(args, capture_output=True)

            if result.stderr:
                print(result.stderr.decode())
            if result.stdout:
                print(result.stdout.decode())

            result.check_returncode()

            bgfx_shader = BgfxShader()
            bgfx_shader.read(f, platform, stage)

        return bgfx_shader


def generate_bgfx_defines(platform: ShaderPlatform, stage: ShaderStage):
    keys = [
        # Platform
        "BX_PLATFORM_ANDROID",
        "BX_PLATFORM_EMSCRIPTEN",
        "BX_PLATFORM_IOS",
        "BX_PLATFORM_LINUX",
        "BX_PLATFORM_OSX",
        "BX_PLATFORM_PS4",
        "BX_PLATFORM_WINDOWS",
        "BX_PLATFORM_XBOXONE",
        # Language
        "BGFX_SHADER_LANGUAGE_GLSL",
        "BGFX_SHADER_LANGUAGE_HLSL",
        "BGFX_SHADER_LANGUAGE_METAL",
        "BGFX_SHADER_LANGUAGE_PSSL",
        "BGFX_SHADER_LANGUAGE_SPIRV",
        # Stage
        "BGFX_SHADER_TYPE_COMPUTE",
        "BGFX_SHADER_TYPE_FRAGMENT",
        "BGFX_SHADER_TYPE_VERTEX",
    ]

    defines = {key: 0 for key in keys}

    # Stage
    if stage == ShaderStage.Compute:
        defines["BGFX_SHADER_TYPE_COMPUTE"] = 1
    elif stage == ShaderStage.Vertex:
        defines["BGFX_SHADER_TYPE_VERTEX"] = 1
    else:
        defines["BGFX_SHADER_TYPE_FRAGMENT"] = 1

    # Platform & language
    if platform == ShaderPlatform.Metal:
        defines["BX_PLATFORM_IOS"] = 1
        defines["BGFX_SHADER_LANGUAGE_METAL"] = 1

    elif platform.name.startswith("ESSL"):
        defines["BX_PLATFORM_ANDROID"] = 1
        defines["BGFX_SHADER_LANGUAGE_GLSL"] = int(platform.name.removeprefix("ESSL_"))

    elif platform.name.startswith("GLSL"):
        defines["BX_PLATFORM_LINUX"] = 1
        defines["BGFX_SHADER_LANGUAGE_GLSL"] = int(platform.name.removeprefix("GLSL_"))

    elif platform.name.startswith("Direct3D"):
        defines["BX_PLATFORM_WINDOWS"] = 1
        defines["BGFX_SHADER_LANGUAGE_HLSL"] = (
            400 if platform == ShaderPlatform.Direct3D_SM40 else 500
        )

    elif platform == ShaderPlatform.PSSL:
        defines["BX_PLATFORM_PS4"] = 1
        defines["BGFX_SHADER_LANGUAGE_PSSL"] = 1

    elif platform == ShaderPlatform.Vulkan:
        defines["BGFX_SHADER_LANGUAGE_SPIRV"] = 1

    return [MacroDefine(key, value) for key, value in defines.items()]


# BX_PLATFORM_ANDROID 0 1
# BX_PLATFORM_EMSCRIPTEN 0 1
# BX_PLATFORM_IOS 0 1
# BX_PLATFORM_LINUX 0 1
# BX_PLATFORM_OSX 0 1
# BX_PLATFORM_PS4 0 1
# BX_PLATFORM_WINDOWS 0 1
# BX_PLATFORM_XBOXONE 0 1

# BGFX_SHADER_LANGUAGE_GLSL (ESSL) id [120] 130 140 150 330 400 410 420 [430] 440 ESSL [100] [300] [310] 320
# BGFX_SHADER_LANGUAGE_HLSL id 300 [400] [500]
# BGFX_SHADER_LANGUAGE_METAL 0 1 for ios id for osx 1000
# BGFX_SHADER_LANGUAGE_PSSL 0 1 1000
# BGFX_SHADER_LANGUAGE_SPIRV 0 1 id [1010] 1311 1411 1512 1613

# BGFX_SHADER_TYPE_COMPUTE 0 1
# BGFX_SHADER_TYPE_FRAGMENT 0 1
# BGFX_SHADER_TYPE_VERTEX 0 1



from enum import Enum
from io import BytesIO


class EncryptionType(Enum):
    NONE = "NONE"
    SIMPLE_PASSPHRASE = "SMPL"
    KEY_PAIR = "KYPR"  # Unsupported

    @classmethod
    def read(cls, file: BytesIO):
        return cls(file.read(4).decode()[::-1])

    def write(self, file: BytesIO):
        file.write(self.value[::-1].encode())



import os, pyjson5
from lazurite.material import Material
from lazurite.material.shader_pass.shader_definition import ShaderPlatform
from lazurite.compiler.macro_define import MacroDefine


class ProjectConfig:
    macros: list[MacroDefine]
    platforms: list[ShaderPlatform]
    merge_source: list[str]
    include_patterns: list[str]
    exclude_patterns: list[str]

    def __init__(self) -> None:
        self.macros = []
        self.platforms = []
        self.merge_source = []
        self.include_patterns = ["*"]
        self.exclude_patterns = [".*", "_*"]

    def read_json_file(self, path: str, profiles: list[str]):
        def append_unique(target: list, source: list):
            """
            Appends elements from source to target, if they aren't in target list already.
            """
            target.extend((item for item in source if item not in target))

        if not os.path.isfile(path):
            return
        with open(path) as f:
            json_data = pyjson5.load(f)
        has_macros = False
        has_platforms = False
        has_merge = False
        has_include_pattern = False
        has_exclude_pattern = False
        if "profiles" in json_data:
            json_profiles = json_data["profiles"]
            for profile in profiles:
                if profile not in json_profiles:
                    print(f'Warning: profile "{profile}" was not found!')
                    continue
                profile = json_profiles[profile]

                if "macros" in profile:
                    if not has_macros:
                        self.macros = []
                        has_macros = True
                    append_unique(
                        self.macros,
                        [MacroDefine.from_string(m) for m in profile["macros"]],
                    )

                if "platforms" in profile:
                    if not has_platforms:
                        self.platforms = []
                        has_platforms = True
                    append_unique(
                        self.platforms,
                        (ShaderPlatform[p] for p in profile["platforms"]),
                    )

                if "merge_source" in profile:
                    if not has_merge:
                        self.merge_source = []
                        has_merge = True
                    append_unique(self.merge_source, profile["merge_source"])

                if "include_patterns" in profile:
                    if not has_include_pattern:
                        self.include_patterns = []
                        has_include_pattern = True
                    patterns = profile["include_patterns"]
                    append_unique(
                        self.include_patterns,
                        [patterns] if type(patterns) == str else patterns,
                    )

                if "exclude_patterns" in profile:
                    if not has_exclude_pattern:
                        self.exclude_patterns = []
                        has_exclude_pattern = True
                    patterns = profile["exclude_patterns"]
                    append_unique(
                        self.exclude_patterns,
                        [patterns] if type(patterns) == str else patterns,
                    )

        if "base_profile" in json_data:
            base_profile = json_data["base_profile"]
            if "macros" in base_profile and not has_macros:
                self.macros = [
                    MacroDefine.from_string(m) for m in base_profile["macros"]
                ]

            if "platforms" in base_profile and not has_platforms:
                self.platforms = [ShaderPlatform[p] for p in base_profile["platforms"]]

            if "merge_source" in base_profile and not has_merge:
                self.merge_source = base_profile["merge_source"]

            if "include_patterns" in base_profile and not has_include_pattern:
                patterns = base_profile["include_patterns"]
                self.include_patterns = (
                    [patterns] if type(patterns) == str else patterns
                )

            if "exclude_patterns" in base_profile and not has_exclude_pattern:
                patterns = base_profile["exclude_patterns"]
                self.exclude_patterns = (
                    [patterns] if type(patterns) == str else patterns
                )

        new_merge_source = []
        for merge_path in self.merge_source:
            merge_path = os.path.normpath(
                os.path.join(os.path.split(path)[0], merge_path)
            )
            if (
                os.path.isfile(merge_path)
                and merge_path not in new_merge_source
                and merge_path.endswith(Material.EXTENSION)
            ):
                new_merge_source.append(merge_path)
            elif os.path.isdir(merge_path):
                for mat_dir in os.scandir(merge_path):
                    if (
                        mat_dir.is_file()
                        and mat_dir.path not in new_merge_source
                        and mat_dir.path.endswith(Material.EXTENSION)
                    ):
                        new_merge_source.append(mat_dir.path)
            else:
                print(f'Warning: merge path "{merge_path}" was not found')
        self.merge_source = new_merge_source



from io import BytesIO

from lazurite import util
from ..platform import ShaderPlatform
from ..stage import ShaderStage
from .bgfx_shader import BgfxShader
from .shader_input import ShaderInput


class ShaderDefinition:
    stage: ShaderStage
    platform: ShaderPlatform
    inputs: list[ShaderInput]
    hash: int
    bgfx_shader: BgfxShader

    def __init__(self) -> None:
        self.stage = ShaderStage.Unknown
        self.platform = ShaderPlatform.Unknown
        self.inputs = []
        self.hash = 0
        self.bgfx_shader = BgfxShader()

    def read(self, file: BytesIO):
        self.stage = ShaderStage[util.read_string(file)]
        self.platform = ShaderPlatform[util.read_string(file)]

        stage_index = util.read_ubyte(file)
        if self.stage.value != stage_index:
            raise Exception(
                f'Stage name "{self.stage.name}" and index "{stage_index}" do not match! Index "{self.stage.value}" was expected.'
            )

        platform_index = util.read_ubyte(file)
        if self.platform.value != platform_index:
            raise Exception(
                f'Platform name "{self.platform.name}" and index "{platform_index}" do not match! Index "{self.platform.value}" was expected.'
            )

        self.inputs = [ShaderInput().read(file) for _ in range(util.read_ushort(file))]
        self.hash = util.read_ulonglong(file)
        bgfx_shader_bytes = BytesIO(util.read_array(file))
        self.bgfx_shader.read(bgfx_shader_bytes, self.platform, self.stage)

        return self

    def write(self, file: BytesIO):
        util.write_string(file, self.stage.name)
        util.write_string(file, self.platform.name)
        util.write_ubyte(file, self.stage.value)
        util.write_ubyte(file, self.platform.value)

        util.write_ushort(file, len(self.inputs))
        for inp in self.inputs:
            inp.write(file)

        util.write_ulonglong(file, self.hash)
        self.bgfx_shader.write(file, self.platform, self.stage)

        return self

    def get_shader_file_name(self, index: int):
        return f"{index}.{self.platform.name}.{self.stage.name}.{self.platform.file_extension()}"

    def serialize_properties(self, index: int):
        obj = {}
        if index != None:
            obj["file_name"] = self.get_shader_file_name(index)
        obj["stage"] = self.stage.name
        obj["platform"] = self.platform.name
        obj["inputs"] = [
            shaderInput.serialize_properties() for shaderInput in self.inputs
        ]
        obj["hash"] = self.hash
        obj["bgfx_shader"] = self.bgfx_shader.serialize_properties()
        return obj

    def load(self, object: dict, path: str):
        self.stage = ShaderStage[object["stage"]]
        self.platform = ShaderPlatform[object["platform"]]
        self.inputs = [ShaderInput().load(inp) for inp in object["inputs"]]
        self.hash = object["hash"]
        self.bgfx_shader = BgfxShader().load(object, path)
        return self

    def label(
        self,
        material_name: str,
        pass_name: str,
        variant_index: int,
        is_supported: bool,
        flags: dict,
    ):
        if not any(
            self.platform.name.startswith(platform_prefix)
            for platform_prefix in ["ESSL", "GLSL", "Metal"]
        ):
            return self

        comment = (
            "// Shader Information:\n"
            f"// - Name: {material_name}\n"
            f"// - Pass: {pass_name}\n"
            f"// - Platform: {self.platform.name}\n"
            f"// - Stage: {self.stage.name}\n"
            f"// - Variant: {variant_index}\n"
            f"// - Variant Supported: {is_supported}\n"
        )

        if flags:
            comment += "// - Variant Flags: \n"
            comment += "\n".join(
                [f"//    - {flag}: {value}" for flag, value in flags.items()]
            )

        code = self.bgfx_shader.shader_bytes.decode()
        code = util.insert_header_comment(code, comment)
        self.bgfx_shader.shader_bytes = code.encode()

        return self



from io import BytesIO
import struct
import re


# Reading binary files.
def read_ulonglong(f: BytesIO) -> int:
    return struct.unpack("<Q", f.read(8))[0]


def read_ulong(f: BytesIO) -> int:
    return struct.unpack("<L", f.read(4))[0]


def read_bool(f: BytesIO) -> bool:
    return struct.unpack("<?", f.read(1))[0]


def read_ubyte(f: BytesIO) -> int:
    return struct.unpack("<B", f.read(1))[0]


def read_ushort(f: BytesIO) -> int:
    return struct.unpack("<H", f.read(2))[0]


def read_array(f: BytesIO) -> bytes:
    return f.read(struct.unpack("<I", f.read(4))[0])


def read_string(f: BytesIO) -> str:
    return read_array(f).decode()


# writing binary files.
def write_ulonglong(f: BytesIO, val: int):
    f.write(struct.pack("<Q", val))


def write_ulong(f: BytesIO, val: int):
    f.write(struct.pack("<L", val))


def write_bool(f: BytesIO, val: bool):
    f.write(struct.pack("<?", val))


def write_ubyte(f: BytesIO, val: int):
    f.write(struct.pack("<B", val))


def write_ushort(f: BytesIO, val: int):
    f.write(struct.pack("<H", val))


def write_array(f: BytesIO, val: bytes):
    f.write(struct.pack("<I", len(val)))
    f.write(val)


def write_string(f: BytesIO, val: str):
    write_array(f, val.encode())


def format_definition_name(name: str):
    # aA -> a_A
    name = re.sub(r"([a-z]+)([A-Z])", r"\1_\2", name)
    # AAa -> A_Aa
    name = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", name)
    # X00 -> X_00
    # name = re.sub(r"([a-zA-Z])(\d+)", r"\1_\2", name)
    # 00X -> 00_X
    name = re.sub(r"(\d+)([a-zA-Z])", r"\1_\2", name)
    return name.upper()


def generate_flag_name_macro(key: str, value: str, is_bool: bool = False):
    if is_bool:
        return format_definition_name(key)
    else:
        return format_definition_name(key + "__" + value)


def generate_pass_name_macro(name: str):
    name = format_definition_name(name)
    if name.endswith("_PASS"):
        return name
    return name + "_PASS"


def insert_header_comment(code: str, comment: str):
    if code.startswith("#version"):
        return code.replace("\n", "\n\n" + comment + "\n\n", 1)
    else:
        return comment + "\n\n" + code



from dataclasses import dataclass
import myers
import sympy
import time
import copy
import io
import re


from lazurite import util


@dataclass
class Variant:
    """
    Decompiler input struct, containing code and flags.
    """

    flags: dict[str, str]
    code: str


class Shader:
    """
    Class, responsible for storing intermediate representations of a shader.
    """

    codes: list[list[int]] | list[str]
    flags: list[set[int]]
    combined_code: list[int]
    line_conditions: list[set[int]]
    chunks: list[tuple[int, int]] | list[tuple[int, int, int]]
    is_struct: bool
    all_flags: set[int]

    def __init__(self) -> None:
        self.codes = []
        self.flags = []
        self.chunks = []
        self.is_struct = False
        self.combined_code = []
        self.line_conditions = []

    def load(
        self,
        variants: list[Variant],
        remove_comments=True,
        preprocess=True,
    ):
        """
        Loads and processes shader variants into codes and flags arrays.
        """
        self.codes = []
        self.flags = []
        for i, variant in enumerate(variants):
            code = variant.code
            if remove_comments:
                code = re.sub(r"//.*\n", "", code)
                code = re.sub(r"/\*.*?\*/", "", code, flags=re.DOTALL)
                code = re.sub(r"\n\n+", "\n", code)
            if preprocess:
                code = _preprocess_shader(code)
            self.codes.append(code)
            self.flags.append({i})

    def uniquify(self):
        """
        Removes duplicate codes and combines their flags, also creates `all_flags` array with all flags.
        """
        codes: list[str] = []
        flags: list[set[int]] = []
        all_flags = set()
        for index, code in enumerate(self.codes):
            try:
                i = codes.index(code)
            except ValueError:
                i = len(codes)
                codes.append(code)
                flags.append(set())
            flags[i].update(self.flags[index])
            all_flags.update(self.flags[index])
        self.all_flags = all_flags
        self.codes = codes
        self.flags = flags

    def encode(self, line_decode_table: dict[str, int]):
        """
        Encodes shader as an array of line indices, while saving line-index mapping in `line_decode_table`.
        """
        for i, code in enumerate(self.codes):
            lines = code.splitlines()
            encoded_shader = []
            for line in lines:
                line_index = line_decode_table.get(line)
                if line_index is None:
                    line_index = len(line_decode_table)
                    line_decode_table[line] = line_index
                encoded_shader.append(line_index)
            self.codes[i] = encoded_shader

    def diff(self):
        """
        Compute combined diff of all shader variants. Populates `combined_code` and `line_conditions`.
        """
        combined_code: list[int] = []
        line_conditions: list[set[int]] = []
        for i, code in enumerate(self.codes):
            diff = myers.diff(combined_code, code)
            combined_code = []
            new_conditions = []
            current_index = 0
            for op, val in diff:
                combined_code.append(val)
                if op == "i":
                    new_conditions.append(copy.copy(self.flags[i]))
                elif op == "r":
                    new_conditions.append(line_conditions[current_index])
                    current_index += 1
                elif op == "k":
                    condition_set = line_conditions[current_index]
                    condition_set.update(self.flags[i])
                    new_conditions.append(condition_set)
                    current_index += 1

            line_conditions = new_conditions
        self.combined_code = combined_code
        self.line_conditions = line_conditions

    def group_lines(self):
        """
        Groups lines with the same condition into chunks, populates `Shader.chunks` list.
        """
        start = 0
        chunks: list[tuple[int, int]] = []
        end = len(self.line_conditions) - 1
        current_condition = self.line_conditions[0]
        for i, condition in enumerate(self.line_conditions):
            if i == end:
                if current_condition != condition:
                    chunks.append((start, i))
                    chunks.append((i, i + 1))
                else:
                    chunks.append((start, i + 1))
            elif current_condition != condition:
                chunks.append((start, i))
                start = i
                current_condition = condition
        self.chunks = chunks

    def generate_calc_list(self, calc_list: list[tuple[set[int], set[int]]]):
        """
        Populates `calc_list` with unique calculations that need to be performed, adds calculation index to each chunk.
        """
        for index, (start, end) in enumerate(self.chunks):
            condition = self.line_conditions[start]
            found = False
            try:
                calc_index = calc_list.index((self.all_flags, condition))
            except ValueError:
                calc_index = len(calc_list)
            else:
                found = True

            if len(condition) == len(self.all_flags):
                calc_index = -1  # This line is always unconditionally present.
            elif not found:
                calc_list.append((self.all_flags, condition))

            self.chunks[index] = (start, end, calc_index)

    def assemble_code(self, calc_final, line_decode_table):
        """
        Creates a final code string with macros.
        """
        string_pipe = io.StringIO()
        for start, end, calc in self.chunks:
            has_condition = calc != -1

            if has_condition:
                string_pipe.write(calc_final[calc] + "\n")

            for i in range(start, end):
                string_pipe.write(line_decode_table[self.combined_code[i]] + "\n")

            if has_condition:
                string_pipe.write("#endif\n")
        string_pipe.seek(0)
        return string_pipe.read()


def _extract_functions(
    code: str, functions: dict[str, tuple[bool, Shader]], flags: set[int]
):
    """
    Extracts functions and structs from shader, storing them as separate individual shaders and replacing them in main shader.
    """
    re_func_start = re.compile(
        r"^[\s]*?([^#\s][\w]+)[\s]+([\w]+)[\s]*\(([^;]*?)\)[\s]*{",
        re.DOTALL | re.MULTILINE,
    )
    func_start = re.search(re_func_start, code)
    modified_text = ""

    # Extract functions.
    while func_start:
        groups = func_start.groups()
        args = groups[2].replace("\n", "")
        name = f"{groups[0]} {groups[1]}({args})"

        modified_text += code[: func_start.start()]

        bracket_balance = 1
        for i in range(func_start.end(), len(code)):
            c = code[i]
            if c == "{":
                bracket_balance += 1
            elif c == "}":
                bracket_balance -= 1

            if not bracket_balance:
                content = code[func_start.end() : i]
                code = code[i + 1 :]
                break
        if bracket_balance:
            break

        func = functions.get(name)

        if not func:
            func = Shader()
            functions[name] = func

        modified_text += f"START_NAME|||{name}|||END_NAME\n"

        func.codes.append(content)
        func.flags.append(copy.copy(flags))

        func_start = re.search(re_func_start, code)

    code = modified_text + code

    # Extract structs.
    match: re.Match
    for match in re.finditer(
        r"^[\s]*?struct[\s]+([\w]+)[\s]*{(.*?)};", code, re.DOTALL | re.MULTILINE
    ):
        name, contents = match.groups()
        name = "struct " + name

        struct = functions.get(name)
        if not struct:
            struct = Shader()
            struct.is_struct = True
            functions[name] = struct

        struct.codes.append(contents)
        struct.flags.append(copy.copy(flags))

        code = code.replace(match.group(), f"START_NAME|||{name}|||END_NAME\n")

    return code


def _evaluate_condition(condition: list[set[bool, bool, int, int]], flag: list[int]):
    value = False
    for c in condition:
        a = flag[c[2]] == c[3]
        if c[0]:
            a = not a
        if c[1]:
            value = value and a
        else:
            value = value or a

    return value


def _calculate_condition_score(
    expr: list[set[bool, bool, int, int]],
    conditions: tuple[set[int], set[int]],
    flags: list[list[int]],
):
    score = 0
    for flag in conditions[0]:
        goal = flag in conditions[1]

        score += 1 if goal == _evaluate_condition(expr, flags[flag]) else -1

    return score


def _increment_expr(expr: list[list[bool, bool, int, int]], flag_def: list[int]):
    for i, val in enumerate(expr):
        if not val[1] and i:
            val[1] = True
            return
        val[1] = False

        if val[3] + 1 < flag_def[val[2]]:
            val[3] += 1
            return
        val[3] = 0

        if not val[0]:
            val[0] = True
            return
        val[0] = False

        if val[2] + 1 < len(flag_def):
            val[2] += 1
            return
        val[2] = 0

    expr.append([False, False, 0, 0])


def _slow_search(
    conditions: tuple[set[int], set[int]],
    flags: list[list[int]],
    flag_def: list[int],
    timeout: float = 10,
):
    """
    Slow brute-force search that scores all possible expression combinations.
    """
    # (bool NOT, bool AND/OR, flag_index, flag_value)

    expression = [[False, False, 0, 0]]
    best_score = -10000
    best_expr = []

    t = time.perf_counter()
    while len(expression) <= len(flag_def) + 5:
        score = _calculate_condition_score(expression, conditions, flags)

        if score > best_score:
            best_score = score
            best_expr = copy.deepcopy(expression)

        if score == len(conditions[0]):
            return True, expression, score

        _increment_expr(expression, flag_def)

        if time.perf_counter() - t > timeout:
            break

    return False, best_expr, best_score


def _fast_search(
    conditions: tuple[set[int], set[int]], flags: list[list[int]], flag_def: list[int]
) -> tuple[bool, list, int]:
    """
    Fast search that calculates best score for each new token.
    """
    # (bool NOT, bool AND/OR, flag_index, flag_value)

    best = -10000
    best_expr = []
    local_best_expr = []
    expression = []
    for _ in range(len(flag_def) + 5):
        expression = copy.deepcopy(local_best_expr)
        expression.append([False, False, 0, 0])

        local_best = -10000
        local_best_expr = []

        for i in range(4):
            if len(expression) == 1 and i >= 2:
                break
            expression[-1][0] = i % 2 == 1
            expression[-1][1] = i / 2 >= 1
            for flag_index, flag in enumerate(flag_def):
                expression[-1][2] = flag_index
                for value in range(flag):
                    expression[-1][3] = value

                    score = _calculate_condition_score(expression, conditions, flags)
                    if score > local_best:
                        local_best = score
                        local_best_expr = copy.deepcopy(expression)

        if local_best > best:
            best = local_best
            best_expr = copy.deepcopy(local_best_expr)
            if best == len(conditions[0]):
                return True, best_expr, best
    return False, best_expr, best


def _convert_to_sympy_expression(
    expression: list[set[bool, bool, int, int]],
    inverse_flag_def: list[tuple[str, list[str]]],
):
    expr = sympy.false
    for e in expression:
        flag, values = inverse_flag_def[e[2]]
        value = values[e[3]]
        is_pass = flag == "pass"
        if is_pass:
            flag = util.generate_pass_name_macro(value)
        else:
            is_bool = False
            # is_bool = set(values) in [
            #     {"On", "Off"},
            #     {"Enabled", "Disabled"},
            # ]
            if flag.startswith("f_"):
                flag = flag.removeprefix("f_")
                flag = util.generate_flag_name_macro(flag, value, is_bool)
            else:
                flag = util.format_definition_name(flag + value)

        a = sympy.symbols(flag)

        # if not is_pass and is_bool and value in {"Off", "Disabled"}:
        #     a = ~a

        if e[0]:
            a = ~a
        if e[1]:
            expr = expr & a
        else:
            expr = expr | a

    return expr


def _format_expression(expr: str):
    """
    Formats stringified sympy expression according to GLSL macro format.
    """
    expr = expr.replace("~", "!").replace("|", "||").replace("&", "&&")
    expr = re.sub(r"(\w+)", r"defined(\1)", expr)
    return expr


def _convert_from_dict_to_list(line_decode_table: dict[str, int]):
    l = [""] * len(line_decode_table)
    for line, i in line_decode_table.items():
        l[i] = line
    return l


def restore_code(
    variants: list[Variant],
    remove_comments=True,
    process_shaders=False,
    search_timeout: float = 10,
) -> tuple[set[str], str]:
    """
    Attempts to restore original shader source, by combining variants while adding missing macros.
    """
    main_shader = Shader()
    main_shader.load(variants, remove_comments, process_shaders)
    main_shader.uniquify()

    functions: dict[str, Shader] = {}
    for i, (code, flags) in enumerate(zip(main_shader.codes, main_shader.flags)):
        main_shader.codes[i] = _extract_functions(code, functions, flags)

    for _, func in functions.items():
        func.uniquify()

    code_flag_list = [copy.copy(v.flags) for v in variants]

    line_decode_table = {}
    main_shader.encode(line_decode_table)

    for _, func in functions.items():
        func.encode(line_decode_table)

    line_decode_table = _convert_from_dict_to_list(line_decode_table)

    # { name: (index, [values]) }
    flag_definition: dict[str, tuple[int, list[str]]] = {}
    for variant in variants:
        for name, value in variant.flags.items():
            definition = flag_definition.get(name)
            if definition is None:
                flag_definition[name] = (len(flag_definition), [value])
            else:
                value_list = definition[1]
                if value not in value_list:
                    value_list.append(value)

    keys_to_remove = []
    for name, (_, value_list) in flag_definition.items():
        # Remove flags with a single value.
        if len(value_list) <= 1:
            keys_to_remove.append(name)
        # Bias against disabling
        for i in {"Off", "Disabled"}:
            if i in value_list:
                value_list.remove(i)
                value_list.append(i)
        # Bias towards enabling
        for i in {"On", "Enabled"}:
            if i in value_list:
                value_list.remove(i)
                value_list.insert(0, i)
    for key in keys_to_remove:
        flag_definition.pop(key)
    for flags in code_flag_list:
        for key in keys_to_remove:
            flags.pop(key)

    # Re-populate flag indices.
    for index, (flag, (_, values)) in enumerate(flag_definition.items()):
        flag_definition[flag] = (index, values)

    main_shader.diff()
    main_shader.group_lines()

    for _, func in functions.items():
        func.diff()
        func.group_lines()

    calc_list: list[tuple[set[int], set[int]]] = []
    main_shader.generate_calc_list(calc_list)
    for _, func in functions.items():
        func.generate_calc_list(calc_list)

    encoded_flags: list[list[int]] = []
    for flags in code_flag_list:
        encoded_flag = [-1] * len(flag_definition)
        for name, value in flags.items():
            name_index, value_list = flag_definition[name]
            encoded_flag[name_index] = value_list.index(value)
        encoded_flags.append(encoded_flag)

    encoded_flag_def = [0] * len(flag_definition)
    for _, (index, values) in flag_definition.items():
        encoded_flag_def[index] = len(values)

    inverse_flag_def: list[tuple[str, list[str]]] = [0] * len(flag_definition)
    for flag, (index, values) in flag_definition.items():
        inverse_flag_def[index] = (flag, values)

    macro_pattern = re.compile(r"\w+")
    macro_set = set()
    calc_result = []
    calc_result_formatted = []
    calc_final = []
    for i, calc in enumerate(calc_list):
        success, expr, score = _fast_search(calc, encoded_flags, encoded_flag_def)
        if not success:
            print("slow search")
            success, expr_slow, score_slow = _slow_search(
                calc, encoded_flags, encoded_flag_def, search_timeout
            )
            if not success:
                print("not found")
            if success or score_slow > score:
                expr = expr_slow
                score = score_slow

        result = _convert_to_sympy_expression(expr, inverse_flag_def)
        try:
            index = calc_result.index(result)
        except ValueError:
            index = len(calc_result)
            calc_result.append(result)
            result = sympy.simplify_logic(result, force=True)
            str_result = str(result)
            macro_set.update(set(re.findall(macro_pattern, str_result)))
            if len(result.atoms()) == 1:
                if str_result.startswith("~"):
                    str_result = "#ifndef " + str_result.removeprefix("~")
                else:
                    str_result = "#ifdef " + str_result
            else:
                str_result = "#if " + _format_expression(str_result)
            if not success:
                str_result = f"// Approximation, matches {score} cases out of {len(calc[0])}\n{str_result}"

            calc_result_formatted.append(str_result)

        calc_final.append(calc_result_formatted[index])

    main_code = main_shader.assemble_code(calc_final, line_decode_table)

    for name, func in functions.items():
        code = func.assemble_code(calc_final, line_decode_table)
        if not code.startswith("\n"):
            code = "\n" + code

        main_code = main_code.replace(
            f"START_NAME|||{name}|||END_NAME",
            f"{name} {{{code}}}" + (";" if func.is_struct else ""),
        )

    if process_shaders:
        main_code = _postprocess_shader(main_code)

    return macro_set, main_code


def _preprocess_shader(shader: str):
    """
    Pre-processes plain text shader code to convert it from GLSL to BGFX SC.\n
    Removes built-in `u_` uniforms, replaces `gl_FragColor` and `gl_FragData`,
    replaces attributes and varyings with `$input` and `$output`, removes macros,
    replaces samplers with BGFX AUTOREG macros, adds NUM_THREADS to compute shaders.
    """
    shader = re.sub(r"^uniform\s+\w+\s+u_[\w[\]]+;\n", "", shader, flags=re.MULTILINE)

    shader = re.sub(r"(\W)bgfx_FragColor(\W)", r"\1gl_FragColor\2", shader)
    shader = re.sub(r"(\W)bgfx_FragData(\W)", r"\1gl_FragData\2", shader)

    shader = re.sub(r"^out\s.+?;\n", "", shader, flags=re.MULTILINE)

    is_vertex_stage = bool(
        re.search(r"^#define varying out$", shader, flags=re.MULTILINE)
    )

    shader = re.sub(r"^#define\s.+?\n", "", shader, flags=re.MULTILINE)
    shader = re.sub(r"^#if\s.+?#endif\n", "", shader, flags=re.MULTILINE | re.DOTALL)
    shader = re.sub(r"^#extension\s.+?\n", "", shader, flags=re.MULTILINE)

    shader = re.sub(
        r"^[\s\w]*?varying\s.+? (\w+);$",
        r"$output \1" if is_vertex_stage else r"$input \1",
        shader,
        flags=re.MULTILINE,
    )
    shader = re.sub(
        r"^[\s\w]*?attribute\s.+? (\w+);$", r"$input \1", shader, flags=re.MULTILINE
    )

    shader = re.sub(r"^#version\s.+?\n", "", shader)

    samplers = [
        (r"lowp sampler2D", r"SAMPLER2D"),
        (r"highp sampler2DMS", r"SAMPLER2DMS"),
        (r"highp sampler3D", r"SAMPLER3D"),
        (r"lowp samplerCube", r"SAMPLERCUBE"),
        (r"highp sampler2DShadow", r"SAMPLER2DSHADOW"),
        (r"highp sampler2D", r"SAMPLER2D_HIGHP"),
        (r"highp samplerCube", r"SAMPLERCUBE_HIGHP"),
        (r"highp sampler2DArray", r"SAMPLER2DARRAY"),
        (r"highp sampler2DMSArray", r"SAMPLER2DMSARRAY"),
        (r"highp samplerCubeArray", r"SAMPLERCUBEARRAY"),
        (r"highp sampler2DArrayShadow", r"SAMPLER2DARRAYSHADOW"),
        (r"highp isampler2D", r"ISAMPLER2D"),
        (r"highp usampler2D", r"USAMPLER2D"),
        (r"highp isampler3D", r"ISAMPLER3D"),
    ]

    for pattern, repl in samplers:
        pattern = r"^uniform " + pattern + r" (\w+);"
        repl = repl + r"_AUTOREG(\1);"
        shader = shader = re.sub(pattern, repl, shader, flags=re.MULTILINE)

    shader = shader = re.sub(
        r"^layout\(std430, .+?\) readonly buffer (\w+) { (\w+) .+? }",
        r"BUFFER_RO_AUTOREG(\1, \2);",
        shader,
        flags=re.MULTILINE,
    )
    shader = shader = re.sub(
        r"^layout\(std430, .+?\) writeonly buffer (\w+) { (\w+) .+? }",
        r"BUFFER_WR_AUTOREG(\1, \2);",
        shader,
        flags=re.MULTILINE,
    )
    shader = shader = re.sub(
        r"^layout\(std430, .+?\) buffer (\w+) { (\w+) .+? }",
        r"BUFFER_RW_AUTOREG(\1, \2)",
        shader,
        flags=re.MULTILINE,
    )

    for i, access in enumerate(("readonly ", "writeonly ", "")):
        for prefix in "iu":
            prefix = "u" if prefix == "u" else ""
            access_id = ("RO", "WR", "RW")[i]

            name = f"{prefix.upper()}IMAGE2D_{access_id}_AUTOREG"
            shader = re.sub(
                r"^layout\((.+?), .+?\) "
                + access
                + "uniform highp "
                + prefix
                + r"image2D (\w+)",
                name + r"(\2, \1)",
                shader,
                flags=re.MULTILINE,
            )

            name = f"{prefix.upper()}IMAGE2D_ARRAY_{access_id}_AUTOREG"
            shader = re.sub(
                r"^layout\((.+?), .+?\) "
                + access
                + "uniform highp "
                + prefix
                + r"image2DArray (\w+)",
                name + r"(\2, \1)",
                shader,
                flags=re.MULTILINE,
            )

            name = f"{prefix.upper()}IMAGE3D_{access_id}_AUTOREG"
            shader = re.sub(
                r"^layout\((.+?), .+?\) "
                + access
                + "uniform highp "
                + prefix
                + r"image3D (\w+)",
                name + r"(\2, \1)",
                shader,
                flags=re.MULTILINE,
            )

    shader = re.sub(
        r"^layout \(local_size_x = (\d+), local_size_y = (\d+), local_size_z = (\d+)\) in;",
        r"NUM_THREADS(\1, \2, \3)",
        shader,
        flags=re.MULTILINE,
    )

    return shader


def _postprocess_shader(shader: str):
    """
    Post-processes plain text shader code to convert it from GLSL to BGFX SC.\n
    Merges `$input` and `$output` declarations together and adds `// Attention!`
    comment to potential array access and matrix multiplication operations.
    """
    shader = shader.splitlines(keepends=True)

    new_shader = []
    args = []
    line_type = 0  # 0 - none, 1 - input, 2 = output
    line_prefix = ""
    for line in shader:
        if line.startswith("$input "):
            current_line_type = 1
            line_prefix = "$input "
        elif line.startswith("$output "):
            current_line_type = 2
            line_prefix = "$output "
        else:
            current_line_type = 0

        if line_type:
            if line_type == current_line_type:
                args.append(line.removeprefix(line_prefix).removesuffix("\n"))
            else:
                new_shader.append(", ".join(args) + "\n")
        if not line_type or line_type != current_line_type:
            if current_line_type:
                args = [line.removesuffix("\n")]
            else:
                new_shader.append(line)

        line_type = current_line_type
    shader = new_shader

    for i, line in enumerate(shader):
        if ") * (" in line or "][" in line:
            line = line[:-1] + " // Attention!\n"
            shader[i] = line

    shader = "".join(shader)

    return shader



import pyjson5, json, os
from Crypto.Cipher import AES
from io import BytesIO

from lazurite.decompiler.macro_decompiler import Variant, restore_code
from lazurite.decompiler.varying_decompiler import (
    restore_varying,
    generate_varying_line,
)
from lazurite import util

from .uniform import Uniform
from .buffer import Buffer
from .shader_pass import Pass
from .platform import ShaderPlatform
from .stage import ShaderStage
from .shader_pass.shader_input import ShaderInput
from .encryption import EncryptionType


class Material:
    MAGIC = 168942106
    EXTENSION = ".material.bin"
    COMPILED_MATERIAL_DEFINITION = "RenderDragon.CompiledMaterialDefinition"
    VERSION = 22

    version: int
    name: str
    encryption: EncryptionType
    parent: str
    buffers: list[Buffer]
    uniforms: list[Uniform]
    passes: list[Pass]

    _encryption_key: bytes
    _encryption_nonce: bytes

    def __init__(self):
        self.version = self.VERSION
        self.name = ""
        self.encryption = EncryptionType.NONE
        self.parent = ""
        self.buffers = []
        self.uniforms = []
        self.passes = []

        self._encryption_key = b""
        self._encryption_nonce = b""

    def read(self, file: BytesIO):
        """
        Loads material definition from a binary file-like object.
        """
        self._validate_magic(file)
        self._validate_definition(file)
        self.version = util.read_ulonglong(file)
        self._validate_version()
        self._decrypt_and_read(file)

    def _validate_magic(self, file: BytesIO):
        if self.MAGIC != util.read_ulonglong(file):
            raise Exception("Failed to match file magic")

    def _validate_definition(self, file: BytesIO):
        if util.read_string(file) != self.COMPILED_MATERIAL_DEFINITION:
            raise Exception("Failed to recognize file as material")

    def _validate_version(self):
        if self.version != self.VERSION:
            raise Exception(f"Unsupported material version: {self.version}")

    def _decrypt_and_read(self, file: BytesIO):
        self.encryption = EncryptionType.read(file)

        if self.encryption == EncryptionType.SIMPLE_PASSPHRASE:
            self._encryption_key = util.read_array(file)
            self._encryption_nonce = util.read_array(file)

            cipher = AES.new(
                self._encryption_key,
                AES.MODE_GCM,
                self._get_truncated_nonce(),
            )
            file = BytesIO(cipher.decrypt(util.read_array(file)))

        elif self.encryption == EncryptionType.KEY_PAIR:
            raise Exception("Huh, how did we even get here?")

        self._read_remaining(file)

    def _read_remaining(self, file: BytesIO):
        self.name = util.read_string(file)
        self._read_parent(file)
        self._read_items(file, Buffer, self.buffers, util.read_ubyte)
        self._read_items(file, Uniform, self.uniforms, util.read_ushort)
        self._read_items(file, Pass, self.passes, util.read_ushort)
        self._validate_magic(file)

    def _read_parent(self, file: BytesIO):
        self.parent = util.read_string(file) if util.read_bool(file) else ""

    def _read_items(self, file: BytesIO, item_type, item_list, read_count):
        count = read_count(file)
        item_list[:] = [item_type().read(file) for _ in range(count)]

    def write(self, file: BytesIO):
        """
        Writes material definition into a binary file-like object.
        """
        util.write_ulonglong(file, self.MAGIC)
        util.write_string(file, self.COMPILED_MATERIAL_DEFINITION)
        util.write_ulonglong(file, self.version)

        self.encryption.write(file)
        if self.encryption == EncryptionType.SIMPLE_PASSPHRASE:
            util.write_array(file, self._encryption_key)
            util.write_array(file, self._encryption_nonce)

            data = BytesIO()
            self._write_remaining(data)
            data.seek(0)
            data = data.read()
            cipher = AES.new(
                self._encryption_key,
                AES.MODE_GCM,
                self._get_truncated_nonce(),
            )
            util.write_array(file, cipher.encrypt(data))

        elif self.encryption == EncryptionType.KEY_PAIR:
            raise Exception("Huh, how did we even get here?")

        else:
            self._write_remaining(file)

    def _write_remaining(self, file: BytesIO):
        util.write_string(file, self.name)

        util.write_bool(file, bool(self.parent))
        if self.parent:
            util.write_string(file, self.parent)

        self._write_items(file, self.buffers, util.write_ubyte)
        self._write_items(file, self.uniforms, util.write_ushort)
        self._write_items(file, self.passes, util.write_ushort)

        util.write_ulonglong(file, self.MAGIC)

    def _write_items(self, file: BytesIO, item_list: list, write_count):
        write_count(file, len(item_list))
        for item in item_list:
            item.write(file)

    def _get_truncated_nonce(self):
        return self._encryption_nonce[:12]

    def serialize_properties(self):
        """
        Returns a dictionary with encoded material properties.
        """
        return {
            "version": self.version,
            "name": self.name,
            "parent": self.parent,
            "buffers": [buffer.name for buffer in self.buffers],
            "uniforms": [uniform.name for uniform in self.uniforms],
            "passes": [render_pass.name for render_pass in self.passes],
        }

    def store(self, name: str, path: str = ".", skip_shaders=False):
        """
        Stores material as a file structure.
        """
        material_dir = os.path.join(path, name)
        os.makedirs(material_dir)

        self._store_json(material_dir)
        self._store_items(material_dir, "buffers", self.buffers)
        self._store_items(material_dir, "uniforms", self.uniforms)

        # That should keep the project protected legally.
        if self.encryption != EncryptionType.NONE:
            print(
                f"Warning! {self.name} material is encrypted. "
                "This tool cannot be used to obtain decrypted shaders."
            )
            skip_shaders = True
        self._store_items(material_dir, "passes", self.passes, skip_shaders)

    def _store_json(self, material_dir):
        json_path = os.path.join(material_dir, "material.json")
        with open(json_path, "w") as f:
            json.dump(self.serialize_properties(), f, indent=4)

    def _store_items(
        self, material_dir: str, subfolder: str, item_list: list, skip_shaders=False
    ):
        if item_list:
            subfolder_dir = os.path.join(material_dir, subfolder)
            os.mkdir(subfolder_dir)

            args = [True] if skip_shaders else []
            for item in item_list:
                item.store(subfolder_dir, *args)

    @classmethod
    def load_bin_file(cls, path: str):
        """
        Creates a material definition from binary file at specified path.
        """
        if os.path.isfile(path):
            material = cls()
            with open(path, "rb") as f:
                material.read(f)
            return material
        else:
            raise Exception(f'Failed to load material at "{path}", it\'s not a file')

    def load_unpacked_material(self, material_path: str):
        """
        Loads unpacked material by merging it with existing material.
        """
        material_json = {}
        material_json_path = os.path.join(material_path, "material.json")
        if os.path.isfile(material_json_path):
            with open(material_json_path) as f:
                material_json: dict = pyjson5.load(f)
        self.version = material_json.get("version", self.VERSION)
        self._validate_version()

        self.name = material_json.get("name", self.name)
        self.parent = material_json.get("parent", self.parent)

        self._load_folder(material_json, material_path, "buffers", self.buffers, Buffer)
        self._load_folder(
            material_json, material_path, "uniforms", self.uniforms, Uniform
        )
        self._load_folder(material_json, material_path, "passes", self.passes, Pass)

    def _load_folder(
        self,
        material_json: dict,
        material_dir: str,
        item_name: str,
        items_list: list,
        constructor,
    ):
        item_names = (
            material_json[item_name]
            if item_name in material_json
            else [item.name for item in items_list]
        )
        folder_path = os.path.join(material_dir, item_name)

        # Read all json files in the folder.
        item_definitions = self._read_item_definitions(folder_path)

        new_items = []
        for name in item_names:
            # Find existing item in the list by name, otherwise create new one.
            item_object = next(
                (item for item in items_list if item.name == name), constructor()
            )
            item_object.name = name
            if name in item_definitions:
                item_object.load(item_definitions[name], folder_path)
            new_items.append(item_object)

        if item_name not in material_json:
            for name, value in item_definitions.items():
                if name in item_names:
                    continue

                new_items.append(constructor().load(value, folder_path))

        items_list[:] = new_items

    def _read_item_definitions(self, folder_path: str):
        item_definitions = {}
        if not os.path.isdir(folder_path):
            return item_definitions
        for file_name in os.listdir(folder_path):
            file_path = os.path.join(folder_path, file_name)
            if not os.path.isfile(file_path) or not file_name.endswith(".json"):
                continue

            with open(file_path) as f:
                definition: dict = pyjson5.load(f)

            if "name" in definition:
                item_definitions[definition["name"]] = definition
            else:
                item_definitions[file_name.removesuffix(".json")] = definition

        return item_definitions

    def label(self):
        """
        Adds comments with information about material into shader code.
        Only for plain text shaders like ESSL, GLSL or Metal.
        """
        for shader_pass in self.passes:
            shader_pass.label(self.name)

    def restore_varying_def(self, search_timeout: float = 10):
        """
        Attempts to restore varying.def.sc file. Works for any platforms.
        """
        permutations: list[Variant] = []
        for p in self.passes:
            per_pass_inputs: dict[
                ShaderPlatform, dict[ShaderStage, list[ShaderInput]]
            ] = {}
            for v in p.variants:
                for s in v.shaders:
                    if s.platform not in per_pass_inputs:
                        per_pass_inputs[s.platform] = {}
                    if s.stage not in per_pass_inputs[s.platform]:
                        per_pass_inputs[s.platform][s.stage] = []
                    input_list = per_pass_inputs[s.platform][s.stage]
                    for i in s.inputs:
                        if i not in input_list:
                            input_list.append(i)

            for platform, stage_dict in per_pass_inputs.items():
                vertex_attributes = []
                fragment_varyings = []
                instance_data = []
                for stage, inputs in stage_dict.items():
                    inputs.sort(key=lambda x: x.name)
                    for i in inputs:
                        is_instance_data, line = generate_varying_line(i, stage)

                        # This shouldn't ever happen.
                        if len([0 for x in inputs if x.name == i.name]) != 1:
                            line += " // ?"

                        if is_instance_data:
                            instance_data.append(line)
                        elif stage == ShaderStage.Vertex:
                            vertex_attributes.append(line)
                        else:
                            fragment_varyings.append(line)
                blocks = []
                if vertex_attributes:
                    blocks.append("\n".join(vertex_attributes))
                if instance_data:
                    blocks.append("\n".join(instance_data))
                if fragment_varyings:
                    blocks.append("\n".join(fragment_varyings))
                if not blocks:
                    continue
                text = "\n\n".join(blocks)
                flags = {"pass": p.name, f"f_platform": platform.name}
                permutations.append(Variant(flags, text))
        if not permutations:
            return ""

        return restore_varying(permutations, search_timeout)

    def restore_shaders(
        self,
        platforms: set[ShaderPlatform],
        stages: set[ShaderStage],
        split_passes=False,
        merge_stages=False,
        process_shaders=False,
        search_timeout: float = 10,
    ) -> list[tuple[ShaderPlatform, ShaderStage, str, str]]:
        """
        Attempts to combine shader permutations into one shader (essl, glsl or metal only only).
        """

        if not self.passes:
            return []

        flag_definition: dict[str, set[str] | list[str]] = {}
        passes: list[str] = []

        restored_shaders: list[tuple[ShaderPlatform, ShaderStage, str, str]] = []

        for p in self.passes:
            passes.append(p.name)
            for key, value in p.default_variant.items():
                if key in flag_definition:
                    flag_definition[key].add(value)
                else:
                    flag_definition[key] = {value}
            for v in p.variants:
                for key, value in v.flags.items():
                    if key in flag_definition:
                        flag_definition[key].add(value)
                    else:
                        flag_definition[key] = {value}
        passes.sort()
        key_list = list(flag_definition.keys())
        key_list.sort()
        flag_definition = {name: flag_definition[name] for name in key_list}
        for name, values in flag_definition.items():
            values = list(values)
            values.sort()
            flag_definition[name] = values

        for platform in platforms:
            shader_definitions: dict[str, dict[ShaderStage, list[Variant]]] = {}
            for shader_pass in self.passes:
                for variant in shader_pass.variants:
                    for shader in variant.shaders:
                        if shader.platform != platform or shader.stage not in stages:
                            continue

                        if shader_pass.name not in shader_definitions:
                            shader_definitions[shader_pass.name] = {}
                        if shader.stage not in shader_definitions[shader_pass.name]:
                            shader_definitions[shader_pass.name][shader.stage] = []
                        code_list = shader_definitions[shader_pass.name][shader.stage]

                        flags = {}
                        if not split_passes:
                            flags["pass"] = shader_pass.name
                        if merge_stages:
                            stage = shader.stage
                            if stage == ShaderStage.Unknown:
                                stage = ShaderStage.Fragment
                            flags["BGFX_SHADER_TYPE_"] = stage.name.upper()
                        for key, value in variant.flags.items():
                            flags["f_" + key] = value

                        code = shader.bgfx_shader.shader_bytes.decode()
                        code_list.append(Variant(flags, code))
            if not shader_definitions:
                continue

            if merge_stages:
                for _, stage_dict in shader_definitions.items():
                    merged_list = []
                    for _, code_list in stage_dict.items():
                        merged_list.extend(code_list)
                    stage_dict.clear()
                    stage_dict[ShaderStage.Fragment] = merged_list
            if not split_passes:
                merged_dict: dict[ShaderStage, list[Variant]] = {}
                for _, stage_dict in shader_definitions.items():
                    for stage, code_list in stage_dict.items():
                        if stage not in merged_dict:
                            merged_dict[stage] = []
                        merged_dict[stage].extend(code_list)
                shader_definitions.clear()
                shader_definitions[self.passes[0].name] = merged_dict

            for shader_pass, stage_dict in shader_definitions.items():
                for stage, code_list in stage_dict.items():
                    macros, code = restore_code(
                        code_list,
                        process_shaders=process_shaders,
                        search_timeout=search_timeout,
                    )
                    # BGFX macros are always defined as either 0 or 1.
                    for stage_name in {"FRAGMENT", "VERTEX", "COMPUTE"}:
                        stage_name = f"BGFX_SHADER_TYPE_{stage_name}"
                        code = (
                            code.replace(f"#ifdef {stage_name}", f"#if {stage_name}")
                            .replace(f"#ifndef {stage_name}", f"#if !{stage_name}")
                            .replace(f"defined({stage_name})", f"{stage_name}")
                        )

                    if flag_definition or passes:
                        comment = "/*\n* Available Macros:"
                        if passes:
                            comment += "\n*\n* Passes:"
                            for pass_name in passes:
                                pass_name = util.generate_pass_name_macro(pass_name)
                                comment += f"\n* - {pass_name}"
                                if pass_name not in macros:
                                    comment += " (not used)"

                        if flag_definition:
                            for flag_name, values in flag_definition.items():
                                comment += f"\n*\n* {flag_name}:"
                                for flag_value in values:
                                    flag = util.generate_flag_name_macro(
                                        flag_name, flag_value, False
                                    )
                                    comment += f"\n* - {flag}"
                                    if flag not in macros:
                                        comment += " (not used)"

                        comment += "\n*/"

                        code = util.insert_header_comment(code, comment)

                        restored_shaders.append((platform, stage, shader_pass, code))

        return restored_shaders

    def sort_variants(self):
        """
        Sorts variants and flags within each variant, useful for diffing.
        """
        for shader_pass in self.passes:
            shader_pass.sort_variants()

    def get_platforms(self):
        """
        Returns available platforms within the material.
        """
        platforms: set[ShaderPlatform] = set()
        for shader_pass in self.passes:
            platforms.update(shader_pass.get_platforms())

        return platforms

    def get_stages(self):
        """
        Returns available stages within the material.
        """
        stages: set[ShaderStage] = set()
        for shader_pass in self.passes:
            stages.update(shader_pass.get_stages())

        return stages

    def merge_variants(self, other: "Material"):
        """
        Merges variants and passes of different materials.
        Useful when creating a single material for multiple platforms.
        """
        for other_pass in other.passes:
            this_pass = next(
                (p for p in self.passes if p.name == other_pass.name), None
            )
            if this_pass is None:
                self.passes.append(other_pass)
            else:
                this_pass.merge_variants(other_pass)

    def get_flag_definitions(self):
        """
        Returns a dict of all possible flag keys and their values.
        """
        definitions: dict[str, set[str]] = {}
        for shader_pass in self.passes:
            pass_defs = shader_pass.get_flag_definitions()
            for key, value in pass_defs.items():
                if key not in definitions:
                    definitions[key] = set()
                definitions[key].update(value)
        return definitions

    def add_platforms(self, platforms: set[ShaderPlatform]):
        for shader_pass in self.passes:
            shader_pass.add_platforms(platforms)

    def remove_platforms(self, platforms: set[ShaderPlatform]):
        for shader_pass in self.passes:
            shader_pass.remove_platforms(platforms)



from enum import Enum


class BlendMode(Enum):
    Unspecified = None
    NoneMode = 0  # None (reserved keyword)
    Replace = 1
    AlphaBlend = 2
    ColorBlendAlphaAdd = 3  # Likely deprecated.
    PreMultiplied = 4
    InvertColor = 5
    Additive = 6
    AdditiveAlpha = 7
    Multiply = 8
    MultiplyBoth = 9
    InverseSrcAlpha = 10
    SrcAlpha = 11
